/*------------------------------------------------------------------------------
 bin/djv_view/file_save.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "app.h"
#include "file.h"
#include <djv_ui/dialog.h>

namespace djv_view {
namespace file {

Save_Info::Save_Info(
  const File & in,
  const File & out,
  const djv_image::Info & info,
  int layer,
  djv_image::data::PROXY proxy,
  const djv_image::Time & time,
  const djv_image::gl_data::Option & option
) :
  in(in),
  out(out),
  info(info),
  layer(layer),
  proxy(proxy),
  time(time),
  option(option)
{}

struct Save::_Save : public djv_ui::Callback_Base
{
  _Save(const Save_Info &, Save * parent);
  ~_Save();
  
private:

  DJV_CALLBACK(_Save, callback, int);
  DJV_CALLBACK(_Save, del_callback, bool);

  Save_Info _info;
  std::auto_ptr<djv_image::io::Load> _load;
  std::auto_ptr<djv_image::io::Save> _save;
  djv_image::Time _load_time, _save_time;
  djv_ui::dialog::Progress * _dialog;
  Save * _parent;
};

Save::_Save::_Save(const Save_Info & in, Save * parent) :
  _info(in),
  _load_time(in.time),
  _save_time(in.out.seq.start(),
    in.out.seq.start() + int(in.time.list.size()) - 1),
  _dialog(0),
  _parent(parent)
{
  using namespace djv_image;
  
  DJV_DBG_OBJ("Save::_Save::_Save");
  DJV_DBG_PRN("in = " << _info.in);
  DJV_DBG_PRN("out = " << _info.out);
  DJV_DBG_PRN("load time = " << _load_time);
  DJV_DBG_PRN("save time = " << _save_time);

  _info.option.xform.rotate = -_info.option.xform.rotate; // XXX WTF?
  const Box2f bbox =
    gl_data::xform_mtx(_info.option.xform) *
    Box2f(_info.info.size * data::proxy_scale(_info.info.proxy));
  DJV_DBG_PRN("bbox = " << bbox);
  _info.option.xform.position = -bbox.position;
  _info.info.size = bbox.size;

  // XXX
    
  const pixel::TYPE type = pixel::type(_info.info.pixel);
  if (type != pixel::F16 && type != pixel::F32)
    _info.option.clamp = true;

  // Open input.
  
  djv_image::Tag tag;
  try
  {
    _load = std::auto_ptr<djv_image::io::Load>(
      djv_image::io::global()->load(_info.in, (Info *)0, &tag));
  }
  catch (Error error)
  {
    app()->error(error, _info.in.name());
    return;
  }

  // Open output.
  
  try
  {
    _save = std::auto_ptr<djv_image::io::Save>(
      djv_image::io::global()->save(_info.out, _info.info, tag, _save_time));
  }
  catch (Error error)
  {
    app()->error(error, _info.out.name());
    return;
  }
  
  // Start...
  
  _dialog = new djv_ui::dialog::Progress;
  _dialog->sig.set(this, callback);
  _dialog->del_sig.set(this, del_callback);
  _dialog->set(_save_time.list.size());
  _dialog->label(str::Format("%%:").arg(_info.out));
  _dialog->show();
}

Save::_Save::~_Save()
{
  DJV_DBG_OBJ("Save::_Save::~_Save");
  
  if (_save.get())
  {
    try
    {
      DJV_DBG_PRN("close");

      _save->close();
    }
    catch (Error error)
    {
      app()->error(error, _info.out.name());
    }
    _save.reset();
  }

  if (_dialog)
  {
    _dialog->sig.del(this);
    _dialog->del_sig.del(this);
    _dialog->del();
  }
}

void Save::_Save::callback(int in)
{
  DJV_DBG_OBJ("Save::_Save::callback");
  DJV_DBG_PRN("in = " << in);

  // Load frame.

  const djv_image::Image * image = 0;
  
  try
  {
    DJV_DBG_PRN("load");

    image = _load->get(_load_time.list[in], _info.layer, _info.proxy);
    DJV_DBG_PRN("image = " << *image);
  }
  catch (Error error)
  {
    app()->error(error, _info.in.name());
    del_callback(false);
    return;
  }
  
  // Process frame.

  djv_image::Image tmp(_info.info);
  tmp.tag = image->tag;
  
  try
  {
    DJV_DBG_PRN("process");
    
    djv_image::gl_data::copy(*image, &tmp, _info.option);
  }
  catch (Error error)
  {
    app()->error(error, _info.in.name());
    del_callback(false);
    return;
  }

  // Save frame.
  
  try
  {
    DJV_DBG_PRN("save");
    
    _save->set(tmp, _save_time.list[in]);
  }
  catch (Error error)
  {
    app()->error(error, _info.out.name());
    del_callback(false);
    return;
  }
}

void Save::_Save::del_callback(bool)
{
  DJV_DBG_OBJ("Save::_Save::del_callback");
  
  _parent->del(this);
}

Save::~Save()
{
  DJV_DBG_OBJ("Save::~Save");
  
  for (uint i = 0; i < _list.size(); ++i) delete _list[i];
}

void Save::set(const Save_Info & in)
{
  DJV_DBG_OBJ("Save::Save::set");
  
  _list += new _Save(in, this);
}

void Save::del(_Save * in)
{
  DJV_DBG_OBJ("Save::Save::del");
  
  for (
    List<_Save *>::iterator i = _list.begin(), end = _list.end();
    i != end;
    ++i
  )
    if (in == *i)
    {
      delete *i;
      _list.erase(i);
      break;
    }
}

}
}

