/*------------------------------------------------------------------------------
 bin/djv_view/image_display.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "image_display.h"
#include <djv_ui/button.h>
#include <djv_ui/color.h>
#include <djv_ui/file.h>
#include <djv_ui/group.h>
#include <djv_ui/label.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/num_float.h>

namespace djv_view {
namespace image_display {

struct Dialog::_Dialog
{
  djv_ui::file::Widget * lut;
  djv_ui::num::Float_Slider * brightness, * contrast, * saturation;
  djv_ui::num::Float_Slider * levels_in [2], * gamma, * levels_out [2],
    * soft_clip;
  djv_ui::button::Push * close;
};

Dialog::Dialog() :
  Base(dialog_title),
  sig(this),
  _obj(0),
  _p(new _Dialog)
{
  using namespace djv_ui;
  
  // Widgets.
  
  group::Box * lut_group = new group::Box(dialog_lut);
  _p->lut = new djv_ui::file::Widget;
  
  group::Box * color_group = new group::Box(dialog_color);
  _p->brightness = new num::Float_Slider;
  _p->brightness->default_value(image::Display().color.brightness);
  _p->brightness->range(0.0, 4.0);
  _p->brightness->label(new label::Widget(dialog_brightness));
  _p->contrast = new num::Float_Slider;
  _p->contrast->default_value(image::Display().color.contrast);
  _p->contrast->range(0.0, 4.0);
  _p->contrast->label(new label::Widget(dialog_contrast));
  _p->saturation = new num::Float_Slider;
  _p->saturation->default_value(image::Display().color.saturation);
  _p->saturation->range(0.0, 4.0);
  _p->saturation->label(new label::Widget(dialog_saturation));
  
  group::Box * levels_group = new group::Box(dialog_levels);
  _p->levels_in[0] = new num::Float_Slider;
  _p->levels_in[0]->default_value(image::Display().levels.in_low);
  _p->levels_in[0]->range(0.0, 1.0);
  _p->levels_in[0]->label(new label::Widget(dialog_levels_in));
  _p->levels_in[1] = new num::Float_Slider;
  _p->levels_in[1]->default_value(image::Display().levels.in_high);
  _p->levels_in[1]->range(0.0, 1.0);
  _p->levels_in[1]->label(new label::Widget);
  _p->gamma = new num::Float_Slider;
  _p->gamma->default_value(image::Display().levels.gamma);
  _p->gamma->range(0.0, 4.0);
  _p->gamma->label(new label::Widget(dialog_gamma));
  _p->levels_out[0] = new num::Float_Slider;
  _p->levels_out[0]->default_value(image::Display().levels.out_low);
  _p->levels_out[0]->range(0.0, 1.0);
  _p->levels_out[0]->label(new label::Widget(dialog_levels_out));
  _p->levels_out[1] = new num::Float_Slider;
  _p->levels_out[1]->default_value(image::Display().levels.out_high);
  _p->levels_out[1]->range(0.0, 1.0);
  _p->levels_out[1]->label(new label::Widget);
  _p->soft_clip = new num::Float_Slider;
  _p->soft_clip->default_value(image::Display().soft_clip);
  _p->soft_clip->range(0.0, 1.0);
  _p->soft_clip->label(new label::Widget(dialog_soft_clip));

  _p->close = new button::Push(djv_ui::dialog::label_close);
  
  // Layout.
  
  layout::group(List<layout::Item *>() << _p->brightness << _p->contrast <<
    _p->saturation << _p->levels_in[0] << _p->levels_in[1] << _p->gamma <<
    _p->levels_out[0] << _p->levels_out[1] << _p->soft_clip);
  
  Layout_V * layout = new Layout_V(this);
  
  layout->add(lut_group);
  Layout_V * layout_group = new Layout_V(lut_group);
  Layout_H * layout_h = new Layout_H(layout_group);
  layout_h->margin(0);
  layout_h->add(_p->lut);
  
  layout->add(color_group);
  color_group->layout()->add(_p->brightness);
  color_group->layout()->add(_p->contrast);
  color_group->layout()->add(_p->saturation);
  
  layout->add(levels_group);
  levels_group->layout()->add(_p->levels_in[0]);
  levels_group->layout()->add(_p->levels_in[1]);
  levels_group->layout()->add(_p->gamma);
  levels_group->layout()->add(_p->levels_out[0]);
  levels_group->layout()->add(_p->levels_out[1]);
  levels_group->layout()->add(_p->soft_clip);
  
  layout->add_spacer(-1, true);
  
  layout_h = new Layout_H(layout);
  layout_h->margin(0);
  layout_h->add_stretch();
  layout_h->add(_p->close);
  
  // Initialize.

  widget_update();
  
  size(size_hint());
  
  // Callbacks.
  
  _p->lut->sig.set(this, lut_callback);
  
  _p->brightness->sig.set(this, brightness_callback);
  _p->contrast->sig.set(this, contrast_callback);
  _p->saturation->sig.set(this, saturation_callback);

  _p->levels_in[0]->sig.set(this, levels_in0_callback);
  _p->levels_in[1]->sig.set(this, levels_in1_callback);
  _p->gamma->sig.set(this, gamma_callback);
  _p->levels_out[0]->sig.set(this, levels_out0_callback);
  _p->levels_out[1]->sig.set(this, levels_out1_callback);
  _p->soft_clip->sig.set(this, soft_clip_callback);
  
  _p->close->sig.set(this, close_callback);
}

Dialog::~Dialog()
{}

void Dialog::set(const void * in, const image::Display & value)
{
  _obj = in;
  _value = value;
  
  widget_update();
}

void Dialog::pick(const void * in, const image::Display & value)
{
  if (in == _obj || ! shown() || ! visible()) return;
  
  set(in, value);
}

void Dialog::update(const void * in, const image::Display & value)
{
  if (in != _obj) return;
  
  _value = value;
  
  widget_update();
}

void Dialog::del(const void * in)
{
  if (in != _obj) return;
  
  _obj = 0;
}

void Dialog::show()
{
  _p->close->take_focus();
  
  djv_ui::dialog::Base::show();
}

void Dialog::widget_callback(const image::Display & in)
{
  _value = in;
  
  widget_update();
}

void Dialog::lut_callback(const File & in)
{
  _value.lut_file = in;
  const djv_image::Image * image = image::lut_load(_value.lut_file);
  _value.lut = image ? *image : djv_image::Image();
  
  sig.emit(_value);
}

void Dialog::brightness_callback(double in)
{
  _value.color.brightness = in;
  
  sig.emit(_value);
}

void Dialog::contrast_callback(double in)
{
  _value.color.contrast = in;
  
  sig.emit(_value);
}

void Dialog::saturation_callback(double in)
{
  _value.color.saturation = in;
  
  sig.emit(_value);
}

void Dialog::levels_in0_callback(double in)
{
  _value.levels.in_low = in;
  
  sig.emit(_value);
}

void Dialog::levels_in1_callback(double in)
{
  _value.levels.in_high = in;
  
  sig.emit(_value);
}

void Dialog::gamma_callback(double in)
{
  _value.levels.gamma = in;
  
  sig.emit(_value);
}

void Dialog::levels_out0_callback(double in)
{
  _value.levels.out_low = in;
  
  sig.emit(_value);
}

void Dialog::levels_out1_callback(double in)
{
  _value.levels.out_high = in;
  
  sig.emit(_value);
}

void Dialog::soft_clip_callback(double in)
{
  _value.soft_clip = in;
  
  sig.emit(_value);
}

void Dialog::close_callback(bool)
{
  hide();
}

void Dialog::widget_update()
{
  callbacks(false);
  
  _p->lut->set(_value.lut_file);
  
  _p->brightness->set(_value.color.brightness);
  _p->contrast->set(_value.color.contrast);
  _p->saturation->set(_value.color.saturation);

  _p->levels_in[0]->set(_value.levels.in_low);
  _p->levels_in[1]->set(_value.levels.in_high);
  _p->gamma->set(_value.levels.gamma);
  _p->levels_out[0]->set(_value.levels.out_low);
  _p->levels_out[1]->set(_value.levels.out_high);
  _p->soft_clip->set(_value.soft_clip);
  
  callbacks(true);
}

const Str
  dialog_title = "Image Display",
  dialog_lut = "Lookup Table",
  dialog_color = "Color",
  dialog_brightness = "Brightness:",
  dialog_contrast = "Contrast:",
  dialog_saturation = "Saturation:",
  dialog_levels = "Levels",
  dialog_levels_in = "Input:",
  dialog_gamma = "Gamma:",
  dialog_levels_out = "Output:",
  dialog_soft_clip = "Soft Clip:"; 

}
}

