/*------------------------------------------------------------------------------
 bin/djv_view/view_widget.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "view.h"
#include <djv_ui/style.h>
#include <djv_image/time.h>
#include <djv_base/math.h>
#include <FL/Fl.H>
#include <FL/gl.h>

namespace djv_view {
namespace view {

Widget::Widget() :
  pick_sig(this),
  scroll_sig(this),
  scroll_value_sig(this),
  file_sig(this),
  _zoom_tmp(0.0),
  _grid(GRID(0)),
  _hud(false),
  _hud_option(true, _HUD_OPTION_SIZE),
  _inside(false),
  _size_min(1) // XXX Fullscreen mode.
{
  text_font(djv_ui::style::global()->font_bold());
}

void Widget::zoom(double in)
{
  djv_ui::image_view::Widget::zoom(in);
}

void Widget::zoom(double in, const Vec2i & focus)
{
  djv_ui::image_view::Widget::zoom(in, focus);
}

double Widget::zoom() const { return djv_ui::image_view::Widget::zoom(); }

void Widget::zoom_focus(double in)
{
  djv_ui::image_view::Widget::zoom(in, _inside ? _mouse : (geom().size / 2));
}

void Widget::grid(GRID in) { _grid = in; }
void Widget::grid_color(const djv_image::Color & in) { _grid_color = in; }

void Widget::hud(bool in) { _hud = in; }
void Widget::hud_info(const Hud & in) { _hud_info = in; }
void Widget::hud_option(const List<bool> & in) { _hud_option = in; }
void Widget::hud_background(HUD_BACKGROUND in)
  { _hud_background = in; }
void Widget::hud_color_foreground(const djv_image::Color & in)
  { _hud_color_foreground = in; }
void Widget::hud_color_background(const djv_image::Color & in)
  { _hud_color_background = in; }

int Widget::handle(int in)
{
  //DJV_DBG_OBJ("view::Widget::handle");
  //DJV_DBG_PRN("in = " << in);
  
  const Vec2i mouse = Vec2i(Fl::event_x(), geom().h - 1 - Fl::event_y());
  switch (in)
  {
    case FL_PUSH:
    {
      take_focus();

      _view_tmp = view();
      _zoom_tmp = zoom();
      
      _mouse = mouse;
      _mouse_td = _mouse;
      
      if (Fl::event_button2() && Fl::event_ctrl())
      {
        zoom(zoom() * 2.0, _mouse);
        redraw();
      }
      else if (Fl::event_button3() && Fl::event_ctrl())
      {
        zoom(zoom() * 0.5, _mouse);
        redraw();
      }
      else if (Fl::event_button2() || Fl::event_ctrl())
      {
        cursor(FL_CURSOR_MOVE);
      }
      else if (Fl::event_button3() || Fl::event_shift())
      {
        cursor(FL_CURSOR_WE);
        scroll_sig.emit(true);
      }
      else
      {
        cursor(FL_CURSOR_CROSS);
        pick_sig.emit(mouse);
      }
    } return 1;
    
    case FL_MOVE: { _mouse = mouse; } return 1;
    
    case FL_DRAG:
    {
      DJV_DBG_OBJ("view::Widget::handle");
    
      _mouse = mouse;
      
      if (Fl::event_button2() || Fl::event_ctrl())
      {
        view(_view_tmp + (_mouse - _mouse_td));
        cursor(FL_CURSOR_MOVE);
        redraw();
      }
      else if (Fl::event_button3() || Fl::event_shift())
      {
        cursor(FL_CURSOR_WE);
        scroll_value_sig.emit((_mouse.x - _mouse_td.x) / 5);
      }
      else
      {
        cursor(FL_CURSOR_CROSS);
        pick_sig.emit(mouse);
      }
    } return 1;
    
    case FL_RELEASE:
      cursor(FL_CURSOR_DEFAULT);
      return 1;
      
    case FL_ENTER: _inside = true; return 1;
    case FL_LEAVE: _inside = false; return 1;
    case FL_FOCUS:
    case FL_UNFOCUS: return 1;
    
    case FL_DND_ENTER:
    case FL_DND_DRAG:
    case FL_DND_LEAVE:
    case FL_DND_RELEASE: return 1;
    case FL_PASTE:
    {
      DJV_DBG_OBJ("view::Widget::handle");
      
      Str tmp(Fl::event_text());
      
      // XXX Convert URL.
      
      static const Str url = "file://";
      if (tmp.find(url) != Str::npos)
        tmp.replace(0, url.size(), "");
      int i = int(tmp.size()) - 1;
      for (; i >= 0; --i)
        if (tmp[i] != '\n' && tmp[i] != '\r') break;
      tmp = Str(tmp, 0, i + 1);
      
      DJV_DBG_PRN("file = " << tmp);
      file_sig.emit(tmp);

    } return 1;
  }
  return djv_ui::image_view::Widget::handle(in);
}

const Vec2i & Widget::mouse() const { return _mouse; }

void Widget::draw()
{
  if (! visible()) return;
  
  DJV_DBG_OBJ("view::Widget::draw");
  
  djv_ui::image_view::Widget::draw();
  if (_grid) draw_grid();
  if (_hud) draw_hud();
}

void Widget::draw_grid()
{
  DJV_DBG_OBJ("view::Widget::draw_grid");
  DJV_DBG_PRN("size = " << geom().size);
  DJV_DBG_PRN("grid = " << label_grid()[_grid]);
  
  int inc = 0;
  switch (_grid)
  {
    case GRID_1x1: inc = 1; break;
    case GRID_10x10: inc = 10; break;
    case GRID_100x100: inc = 100; break;
    default: break;
  }
  
  // Bail if too small.
  
  if ((inc * zoom()) <= 2) return;
  
  // Compute view area.
  
  const Box2i area = Box2i(
    vec::floor<double, int>(Vec2f(-view()) / zoom() / double(inc)) - 1,
    vec::ceil<double, int>(Vec2f(geom().size) / zoom() / double(inc)) + 2) *
    inc;
  DJV_DBG_PRN("area = " << area);
  
  // Draw.
  
  djv_image::Color color(djv_image::pixel::RGB_F32);
  djv_image::color::convert(_grid_color, &color);
  glColor3f(color.get_f32(0), color.get_f32(1), color.get_f32(2));
  
  glPushMatrix();
  glTranslated(view().x, view().y, 0);
  glScaled(zoom(), zoom(), 1.0);
  
  glBegin(GL_LINES);
  for (int y = 0; y <= area.h; y += inc)
  {
    glVertex2i(area.x, area.y + y);
    glVertex2i(area.x + area.w, area.y + y);
  }
  for (int x = 0; x <= area.w; x += inc)
  {
    glVertex2i(area.x + x, area.y);
    glVertex2i(area.x + x, area.y + area.h);
  }
  glEnd();
  
  glPopMatrix();
}

void Widget::draw_hud()
{
  //DJV_DBG_OBJ("view::Widget::draw_hud");
  
  const Box2i geom = this->geom();
  const int margin = djv_ui::style::global()->margin_widget();
  
  // XXX OMGWTFBBQ? This somehow resets the OpenGL state so FLTK's text drawing
  // functions work?
  
  glGetError();

  // Setup.

  //glViewport(0, 0, geom.w, geom.h);
  //image::gl::ortho(Vec2i(geom.w, geom.h));

  gl_font(text_font(), djv_ui::style::global()->font_size());
  
  Vec2i p;
  const int h = gl_height();
  Str tmp;
  
  // Draw.

  p = Vec2i(margin, geom.h - h - margin);
  
  if (_hud_option[HUD_NAME])
  {
    tmp = str::Format("Name = %%").arg(_hud_info.info.name);
    draw_hud(tmp, p, _hud_background, _hud_color_foreground,
      _hud_color_background);
    p.y -= h;
  }

  if (_hud_option[HUD_SIZE])
  {
    tmp = str::Format("Size = %%x%%:%%").arg(_hud_info.info.size.x).
      arg(_hud_info.info.size.y).arg(vec::aspect(_hud_info.info.size), 0, -2);
    draw_hud(tmp, p, _hud_background, _hud_color_foreground,
      _hud_color_background);
    p.y -= h;
  }
  
  if (_hud_option[HUD_PROXY])
  {
    tmp = str::Format("Proxy = %%").arg(_hud_info.info.proxy);
    draw_hud(tmp, p, _hud_background, _hud_color_foreground,
      _hud_color_background);
    p.y -= h;
  }
  
  if (_hud_option[HUD_PIXEL])
  {
    tmp = str::Format("Pixel = %%").
      arg(djv_image::pixel::label_pixel()[_hud_info.info.pixel]);
    draw_hud(tmp, p, _hud_background, _hud_color_foreground,
      _hud_color_background);
  }

  if (_hud_option[HUD_TAG])
  {
    List<Str> tag;
    const List<Str> key_list = _hud_info.tag.key_list();
    for (uint i = 0; i < key_list.size(); ++i)
      tag += str::Format("Tag %% = %%").arg(key_list[i]).
        arg(_hud_info.tag[key_list[i]]);
    
    p = Vec2i(margin, h * (int(tag.size()) - 1) + margin);

    for (List<Str>::const_iterator i = tag.begin(); i != tag.end(); ++i)
    {
      draw_hud(*i, p, _hud_background, _hud_color_foreground,
      _hud_color_background);
      p.y -= h;
    }
  }

  p = Vec2i(geom.w - margin, geom.h - h - margin);
  
  if (_hud_option[HUD_FRAME])
  {
    tmp = str::Format("Frame = %%").
      arg(djv_image::time::units_to_str(_hud_info.frame, _hud_info.speed));
    draw_hud(tmp, Vec2i(p.x - math::ceil(gl_width(tmp.c_str())), p.y),
      _hud_background, _hud_color_foreground, _hud_color_background);
    p.y -= h;
  }
  
  if (_hud_option[HUD_SPEED])
  {
    tmp = str::Format("%%").arg(_hud_info.speed_real, 0, -2);
    p.x -= math::ceil(gl_width(tmp.c_str()));
    draw_hud(tmp, p, _hud_background, _hud_info.dropped_frames ?
      djv_ui::style::global()->color_value().select : _hud_color_foreground,
      _hud_color_background);
    tmp = str::Format("Speed = %%/").arg(
      djv_image::time::speed_to_float(_hud_info.speed), 0, -2);
    p.x -= math::ceil(gl_width(tmp.c_str()));
    draw_hud(tmp, p, _hud_background, _hud_color_foreground,
      _hud_color_background);
  }
}

void Widget::draw_hud(
  const Str & in,
  const Vec2i & position,
  HUD_BACKGROUND background,
  const djv_image::Color & color_foreground,
  const djv_image::Color & color_background
) {
  const int d = gl_descent();
  
  // Draw background.
  
  djv_image::Color _background(djv_image::pixel::RGB_F32);
  djv_image::color::convert(color_background, &_background);
  glColor3f(
    _background.get_f32(0),
    _background.get_f32(1),
    _background.get_f32(2));
    
  switch (background)
  {
    case HUD_BACKGROUND_NONE: break;

    case HUD_BACKGROUND_SOLID:
    {
      const Box2i box(
        box::border(Box2i(position, djv_ui::base::str_size(in)),
        Vec2i(0, 0)));
      glBegin(GL_QUADS);
      glVertex2i(box.x, box.y);
      glVertex2i(box.x + box.w, box.y);
      glVertex2i(box.x + box.w, box.y + box.h);
      glVertex2i(box.x, box.y + box.h);
      glEnd();
    } break;

    case HUD_BACKGROUND_SHADOW:
      gl_draw(in.c_str(), position.x + 1, position.y + d - 1);
      break;
  }

  // Draw foreground.
  
  djv_image::Color _foreground(djv_image::pixel::RGB_F32);
  djv_image::color::convert(color_foreground, &_foreground);
  glColor3f(
    _foreground.get_f32(0),
    _foreground.get_f32(1),
    _foreground.get_f32(2));
  gl_draw(in.c_str(), position.x, position.y + d);
}

const Vec2i & Widget::size_hint() const { return _size_min; }
const Vec2i & Widget::size_min() const { return _size_min; }

}
}

