/*------------------------------------------------------------------------------
 include/djv_image/data.h
 
 Image Data
 ----------
 
 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_DATA_H
#define DJV_IMAGE_DATA_H

#include <djv_image/pixel.h>
#include <djv_base/vec.h>

namespace djv_image {
namespace data {

//------------------------------------------------------------------------------
// Options

enum PROXY
{
  PROXY_NONE,
  PROXY_1_2,
  PROXY_1_4,
  PROXY_1_8,
  _PROXY_SIZE
};

//------------------------------------------------------------------------------
// Information

class Info
{
public:

  Info();
  Info(const Vec2i & size, PIXEL);
  Info(const Str & name, const Vec2i & size, PIXEL);

  Str name;
  Vec2i size;
  PROXY proxy;
  PIXEL pixel;
  bool bgr;
  Vec2b mirror;
  int align;
  mem::ENDIAN endian;

private:

  void init();
};

//------------------------------------------------------------------------------
// Data

class Data
{
public:

  Data();
  Data(const Data &);
  Data(const Info &, const uint8_t * = 0);
  virtual ~Data();
  Data & operator = (const Data &);

  // Value.
  
  void set(const Data &);
  void set(const Info &, const uint8_t * = 0);
  inline const Info & info() const;
  inline const Vec2i & size() const;
  inline PROXY proxy() const;
  inline int w() const;
  inline int h() const;
  inline PIXEL pixel() const;
  inline int channels() const;
  
  // Functions.
  
  void zero();
  
  // Data.
  
  inline uint8_t * data();
  inline const uint8_t * data() const;
  inline uint8_t * data(int x, int y);
  inline const uint8_t * data(int x, int y) const;
  
  inline size_t bytes_pixel() const;
  inline size_t bytes_scanline() const;
  inline size_t bytes_data() const;

private:

  void init();
  void alloc();
  void del();
  
  Info _info;
  int _channels;
  mem::Buffer<uint8_t> _data;
  size_t _bytes_pixel, _bytes_scanline, _bytes_data;
  const uint8_t * _p;
};

//------------------------------------------------------------------------------
// Functions

// Proxy scale.

void proxy_scale(const Data &, Data *, PROXY);
int proxy_scale(PROXY);
Vec2i proxy_scale(const Vec2i &, PROXY);

// Data.

size_t bytes_scanline(const Info &);
size_t bytes_data(const Info &);

void planar_interleave(const Data &, Data *, PROXY = PROXY_NONE);
void planar_deinterleave(const Data &, Data *);

// Gradient.

void gradient(Data *);

//------------------------------------------------------------------------------
// Comparison

bool operator == (const Info &, const Info &);
bool operator != (const Info &, const Info &);

//------------------------------------------------------------------------------
// Conversion

Str & operator >> (Str &, PROXY &) throw (Str);
Str & operator << (Str &, PROXY);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_proxy();

//------------------------------------------------------------------------------
// Debugging

Dbg & operator << (Dbg &, PROXY);
Dbg & operator << (Dbg &, const Info &);
Dbg & operator << (Dbg &, const Data &);

}

//------------------------------------------------------------------------------
// Declarations

using data::Info;
using data::Data;

}

#include <djv_image/data_.h>

#endif

