/*------------------------------------------------------------------------------
 include/djv_image/io_dpx.h

 Image I/O - DPX
 ---------------
 
 SMPTE Digital Picture Exchange image file format.

 DPX is the successor to the Cineon file format, with support for additional
 image and meta data.

 Support:
 
 * Image data:
   * RGB, 10-bit, type "A" packing (the most common variety).
   * L, RGB, RGBA, 8-bit, 16-bit.
 * Interleaved channels only.
 
 References:
 
 * SMPTE, "SMPTE 268M-2003", http:://www.smpte.org
 * Cinesite, "Conversion of 10-bit Log Film Data To 8-bit Linear or Video Data"

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_IO_DPX_H
#define DJV_IMAGE_IO_DPX_H

#include <djv_image/io_cineon.h>

namespace djv_image {
namespace io_dpx {

//------------------------------------------------------------------------------
// Information

static const Str name = "DPX";

static const List<Str> extension_list = List<Str>() << ".dpx";

//------------------------------------------------------------------------------
// Image Tags

enum TAG
{
  TAG_SOURCE_OFFSET,
  TAG_SOURCE_CENTER,
  TAG_SOURCE_SIZE,
  TAG_SOURCE_FILE,
  TAG_SOURCE_TIME,
  TAG_SOURCE_INPUT_DEVICE,
  TAG_SOURCE_INPUT_SERIAL,
  TAG_SOURCE_BORDER,
  TAG_SOURCE_PIXEL_ASPECT,
  TAG_SOURCE_SCAN_SIZE,
  TAG_FILM_FORMAT,
  TAG_FILM_FRAME,
  TAG_FILM_SEQUENCE,
  TAG_FILM_HOLD,
  TAG_FILM_FRAME_RATE,
  TAG_FILM_SHUTTER,
  TAG_FILM_FRAME_ID,
  TAG_FILM_SLATE,
  TAG_TV_INTERLACE,
  TAG_TV_FIELD,
  TAG_TV_VIDEO_SIGNAL,
  TAG_TV_SAMPLE_RATE,
  TAG_TV_FRAME_RATE,
  TAG_TV_TIME_OFFSET,
  TAG_TV_GAMMA,
  TAG_TV_BLACK_LEVEL,
  TAG_TV_BLACK_GAIN,
  TAG_TV_BREAK_POINT,
  TAG_TV_WHITE_LEVEL,
  TAG_TV_INTEGRATION_TIMES
};

//------------------------------------------------------------------------------
// Options

enum TYPE
{
  TYPE_AUTO,
  U10
};

enum ENDIAN
{
  ENDIAN_AUTO,
  MSB,
  LSB
};

enum LOAD_OPTION
{
  LOAD_COLOR,
  LOAD_PRINT
};

struct Load_Option
{
  Load_Option() :
    color(io_cineon::COLOR_AUTO)
  {}

  io_cineon::COLOR color;
  io_cineon::Unprint print;
};

enum SAVE_OPTION
{
  SAVE_COLOR,
  SAVE_PRINT,
  SAVE_TYPE,
  SAVE_ENDIAN
};

struct Save_Option
{
  Save_Option() :
    color(io_cineon::COLOR_PRINT),
    type(U10),
    endian(MSB)
  {}

  io_cineon::COLOR color;
  io_cineon::Print print;
  TYPE type;
  ENDIAN endian;
};

//------------------------------------------------------------------------------
// Load Plugin

class Load : public io::Load
{
public:

  Load();

  Plugin * copy() const;
  const Str & name() const;
  const List<Str> & extension_list() const;
  
  bool option(const Str &, Str *);
  Str option(const Str &) const;
  const List<Str> & option_list() const;

  void open(const File &, Info *, Tag *, Time *) throw (Error);
  void open(const File &, List<Info> *, Tag *, Time *) throw (Error);
  const Image * get(int frame, int layer, data::PROXY) throw (Error);

private:

  void _open(const File &, Info *, Tag *, time::Speed * = 0) throw (Error);
  
  Load_Option _option;
  bool _print;
  std::auto_ptr<Data> _print_lut;
  File _file;
  djv_base::Io _io;
  Image _image;
  Data _tmp;
  Cache _cache;
};

//------------------------------------------------------------------------------
// Save Plugin

class Save : public io::Save
{
public:
  
  Plugin * copy() const;
  const Str & name() const;
  const List<Str> & extension_list() const;
  
  bool option(const Str &, Str *);
  Str option(const Str &) const;
  const List<Str> & option_list() const;

  void open(const File &, const Info &, const Tag &, const Time &)
    throw (Error);
  void set(const Image &, int frame) throw (Error);

private:

  void _open(const File &, const Tag &) throw (Error);

  Save_Option _option;
  File _file;
  djv_base::Io _io;
  Info _info;
  Image _image;
};

//------------------------------------------------------------------------------
// Functions

void info_load(
  djv_base::Io *,
  Info *,
  Tag *,
  time::Speed *,
  bool * print = 0
) throw (Error);

void info_save(
  djv_base::Io *,
  const Str & file,
  const Info &,
  const Tag &,
  ENDIAN,
  io_cineon::COLOR
) throw (Error);

void info_update(djv_base::Io *) throw (Error);

//------------------------------------------------------------------------------
// Conversion

Str & operator >> (Str &, TYPE &) throw (Str);
Str & operator >> (Str &, ENDIAN &) throw (Str);
Str & operator << (Str &, TYPE);
Str & operator << (Str &, ENDIAN);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_type();
const List<Str> & label_endian();
const List<Str> & label_tag();
const List<Str> & label_load_option();
const List<Str> & label_save_option();

}
}

#endif

