/*------------------------------------------------------------------------------
 lib/djv_base/file.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include <djv_base/file.h>
#include <djv_base/math.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>

namespace djv_base {
namespace file {

void File::init()
{
  type = FILE;
  size = user = perm = time = 0;
}

File::File()
{
  init();
}

File::File(const Str & in, bool stat)
{
  init();
  set(in, stat);
}

void File::set(const Str & in, bool stat)
{
  DJV_DBG_OBJ("File::set");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("stat = " << stat);

  init();
  
  // Split file name into pieces.
  
  split(in, &path, &base, &number, &extension);
  DJV_DBG_PRN("path = " << path);
  DJV_DBG_PRN("base = " << base);
  DJV_DBG_PRN("number = " << number);
  DJV_DBG_PRN("extension = " << extension);

  // Information.
  
  if (stat) this->stat();
  
  // Sequence.
  
  seq = str_to_seq(number);
  DJV_DBG_PRN("seq = " << seq);
}

Str File::get(int frame, bool path) const
{
  Str out;
  
  // Reconstruct pieces.

  if (path)
    out += this->path;
  out += base;
  
  if (SEQ == type && frame != -1)
    out += frame_to_str(frame, seq.pad);
  else if (SEQ == type && seq.list.size())
    out += seq_to_str(seq);
  else
    out += number;
  
  out += extension;

  return out;
}

Str File::name() const
{
  return base + number + extension;
}

#if defined(WIN32)
#define _STAT struct ::_stati64
#define _STAT_FNC ::_stati64
#elif defined(DJV_MACOSX) // XXX No stat64?
#define _STAT struct ::stat
#define _STAT_FNC ::stat
#else
#define _STAT struct ::stat64
#define _STAT_FNC ::stat64
#endif

bool File::stat(const Str & path)
{
  DJV_DBG_OBJ("File::stat");
  DJV_DBG_PRN("path = " << path);
  
  _STAT info;
  mem::set(0, &info, sizeof(_STAT));

  const Str file = path_fix(path.size() ? path : this->path) + get(-1, false);
  DJV_DBG_PRN("file = " << file);

  // XXX Windows stat.

  if (0 == _STAT_FNC(file.c_str(), &info)) ;
  else if (0 == _STAT_FNC(Str(file + path_separator).c_str(), &info)) ;
  else if (0 == _STAT_FNC(Str(file, int(file.size()) - 1).c_str(), &info)) ;
  else
  {
    const Str err(strerror(errno));
    DJV_DBG_PRN("errno = " << err);
    return false;
  }

  size = info.st_size;
  user = info.st_uid;
  perm = 0;
  time = info.st_mtime;
  
#if defined(WIN32)
  if (info.st_mode & _S_IFDIR) type = DIR;
  perm |= (info.st_mode & _S_IREAD) ? READ : 0;
  perm |= (info.st_mode & _S_IWRITE) ? WRITE : 0;
  perm |= (info.st_mode & _S_IEXEC) ? EXEC : 0;
#else
  if (S_ISDIR(info.st_mode)) type = DIR;
  if (info.st_mode & S_IRUSR) perm |= READ;
  if (info.st_mode & S_IWUSR) perm |= WRITE;
  if (info.st_mode & S_IXUSR) perm |= EXEC;
#endif
  
  return true;
}

}
}

