/*------------------------------------------------------------------------------
 lib/djv_base/file_seq.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include <djv_base/file.h>
#include <algorithm>

namespace djv_base {
namespace file {

Seq::Seq() :
  pad(0)
{}

Seq::Seq(const List<int> & in, int pad) :
  list(in),
  pad(pad)
{}

Seq::Seq(int start, int end, int pad) :
  pad(pad)
{
  set(start, end);
}

void Seq::set(int start, int end)
{
  if (start < end)
  {
    list.resize(end - start + 1);
    for (int i = start, j = 0; i <= end; ++i, ++j) list[j] = i;
  }
  else
  {
    list.resize(start - end + 1);
    for (int i = start, j = 0; i >= end; --i, ++j) list[j] = i;
  }
}

void seq_sort(Seq * in)
{
  std::sort(in->list.begin(), in->list.end());
}

void seq_compress(List<File> * in, SEQ_COMPRESS compress)
{
  if (SEQ_OFF == compress) return;

  DJV_DBG_OBJ("seq_compress");
  DJV_DBG_PRN("size = " << in->size());
  
  List<File> & list = *in;
  int cache = -1;
  uint i = 0;
  for (uint j = 0; j < list.size(); ++j)
  {
    DJV_DBG_PRN("item = " << list[j]);
    
    const bool seq = list[j].seq_valid();
    if (seq)
    {
      if (cache != -1)
        if (list[cache].seq_add(list[j])) continue;
      
      cache = -1;
      for (uint k = 0; k < i; ++k)
      {
        if (SEQ == list[k].type)
          if (list[k].seq_add(list[j]))
          {
            cache = k;
            continue;
          }
      }
    }
    
    if (! seq || -1 == cache)
    {
      list[i] = list[j];
      if (seq) list[i].type = SEQ;
      ++i;
    }
  }
  DJV_DBG_PRN("size = " << i);
  list.resize(i);
  
  for (uint i = 0; i < list.size(); ++i)
    seq_sort(&list[i].seq);
  
  if (SEQ_RANGE == compress)
    for (uint i = 0; i < list.size(); ++i)
      list[i].seq.set(list[i].seq.start(), list[i].seq.end());
}

List<Str> seq_expand(const File & in)
{
  const uint size = in.seq.list.size();
  if (! size) return List<Str>() << in;

  List<Str> out(size);
  for (uint i = 0; i < size; ++i)
    out[i] = in.get(in.seq.list[i]);
  return out;
}

int seq_find(int in, const List<int> & list)
{
  const uint size = list.size();
  if (! size) return -1;
  
  int out = 0;
  
  int min = 0;
  for (uint i = 0; i < size; ++i)
  {
    const int tmp = math::abs(in - list[i]);
    if (tmp < min || ! i)
    {
      out = i;
      min = tmp;
    }
  }
  
  return out;
}

File seq_match(const File & in, const List<File> & list)
{
  for (uint i = 0; i < list.size(); ++i)
  {
    const File & tmp = list[i];
    if (
      in.base == tmp.base &&
      in.extension == tmp.extension //&&
      //in.pad == tmp.pad
    )
      return tmp;
  }
  return in;
}

namespace {
  
enum INC
{
  BREAK,
  ASCEND,
  DESCEND
};

inline INC inc(int a, int b)
{
  if (a + 1 == b) return ASCEND;
  else if (a - 1 == b) return DESCEND;
  return BREAK;
}

}

Str seq_to_str(const Seq & in)
{
  DJV_DBG_OBJ("seq_to_str");
  DJV_DBG_PRN("in = " << in.list);
  
  List<Str> out;
  
  List<int> list = in.list;
  const uint size = list.size();
  const int pad = in.pad;
  
  // List end marker.
  
  list += -2;

  // Sequence.
  
  int tmp = list[0];
  INC _inc = inc(tmp, list[1]);
  for (uint i = 0; i < size; ++i)
  {
    const INC inc_ = inc(list[i], list[i + 1]);
    DJV_DBG_PRN(list[i] << " = " << inc_);

    if ((inc_ != _inc && _inc != BREAK) || BREAK == inc_)
    {
      if (tmp != list[i])
        out += frame_to_str(tmp, pad) + "-" + frame_to_str(list[i], pad);
      else
        out += frame_to_str(list[i], pad);
      tmp = list[i + 1];
      _inc = inc_;
    }
  }
  
  DJV_DBG_PRN("out = " << out);
  return str::join(out, ',');
}

Seq str_to_seq(const Str & in)
{
  DJV_DBG_OBJ("str_to_seq");
  DJV_DBG_PRN("in = " << in);

  Seq out;
  int pad = 0;
  
  const List<Str> tmp = str::split(in, ',');
  for (uint i = 0; i < tmp.size(); ++i)
  {
    const List<Str> tmp2 = str::split(tmp[i], '-');
    const uint size = tmp2.size();
    if (size)
    {
      int _pad = 0;
      const int start = str_to_frame(tmp2[0], &_pad);
      const int end = str_to_frame(tmp2[size - 1]);
      if (start != -1 && end != -1)
        if (start < end)
          for (int j = start; j <= end; ++j)
            out.list += j;
        else
          for (int j = start; j >= end; --j)
            out.list += j;
      pad = math::max(_pad, pad);
    }
  }

  out.pad = pad;
  DJV_DBG_PRN("out = " << out);
  DJV_DBG_PRN("pad = " << out.pad);
  return out;
}

}
}

