/*------------------------------------------------------------------------------
 lib/djv_base/plugin.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include <djv_base/app.h>
#include <djv_base/dir.h>
#include <djv_base/file.h>
#include <djv_base/plugin.h>
#include <algorithm>
#if ! defined(WIN32)
#include <dlfcn.h>
#endif

namespace djv_base {
namespace plugin {

Plugin::~Plugin()
{}

class Factory::Handle
{
public:

  Handle() :
    _p(0)
  {}

  ~Handle()
  {
#if defined(WIN32)
    if (_p) ::FreeLibrary(_p);
#else
#if ! defined(DJV_IRIX) // XXX IRIX dlopen/pthread bug.
    if (_p) ::dlclose(_p);
#endif
#endif
  }
  
  void open(const Str & in) throw (Error)
  {
#if defined(WIN32)
    _p = ::LoadLibrary(in.c_str());
#else
    _p = ::dlopen(in.c_str(), RTLD_LAZY);
#endif

    if (! _p)
    {
      Str error;
#if ! defined(WIN32)
      error = ::dlerror();
#endif
      throw Error(error);
    }
  }
    
  void * fnc(const Str & in) const
  {
#if defined(WIN32)
    return (void *)::GetProcAddress(_p, in.c_str());
#else
    return ::dlsym(_p, in.c_str());
#endif
  }
  
private:
  
#if defined(WIN32)
  HINSTANCE _p;
#else
  void * _p;
#endif
};

Factory::Factory()
{
  DJV_DBG_OBJ("Factory::Factory");
}

Factory::~Factory()
{
  DJV_DBG_OBJ("Factory::~Factory");
  
  del();
}

void Factory::set(const List<Str> & search_path_list)
{
  //return;

  DJV_DBG_OBJ("Factory::set");
  DJV_DBG_PRN("search paths = " << search_path_list);
  DJV_DBG_PRN("plugin entry = " << plugin_entry());

  //del();

  // XXX OS specific library names.

  Str filter;
//#if defined(DJV_MACOSX)
//  filter = "lib" + plugin_prefix() + "*.dylib";
//#elif defined(WIN32)
#if defined(WIN32)
  filter = "lib" + plugin_prefix() + "*.dll";
#else
  filter = "lib" + plugin_prefix() + "*.so";
#endif
  DJV_DBG_PRN("filter = " << filter);

  // Find plugins.
 
  List<Str> list;
  for (uint i = 0; i < search_path_list.size(); ++i)
  {
    const Dir dir(search_path_list[i]);
    DJV_DBG_PRN("searching = " << dir.get());
    
    List<File> file_list = dir.list();
    file::filter(&file_list, file::FILTER_NONE, List<Str>() << filter);
    for (uint i = 0; i < file_list.size(); ++i)
    {
      DJV_DBG_PRN("found = " << file_list[i]);
      list += file_list[i];
    }
  }
  
  // Load plugins.

  //DJV_DBG_PRN("size = " << list.size());
  for (uint i = 0; i < list.size(); ++i)
  {
    DJV_DBG_PRN("loading = " << list[i]);
    
    // Open.

    std::auto_ptr<Handle> handle(new Handle);
    
    try
    {
      handle->open(list[i]);
    }
    catch (const Error in)
    {
      DJV_DBG_PRN("plugin error = " << in);
      continue;
    }
    
    Plugin_Entry * entry = (Plugin_Entry *)handle->fnc(plugin_entry());
    if (! entry)
    {
      DJV_DBG_PRN("plugin entry point error");
      continue;
    }

    std::auto_ptr<Plugin> plugin;
    try
    {
      plugin = std::auto_ptr<Plugin>(entry());
    }
    catch (const Error in)
    {
      plugin.reset();
      app::global()->error(in);
    }
    if (! plugin.get())
    {
      DJV_DBG_PRN("plugin create error");
      continue;
    }
    DJV_DBG_PRN("name = " << plugin->name());

    // Check for duplicates.
    
    uint j = 0;
    for (; j < _list.size(); ++j)
      if (plugin->name() == _list[j]->name()) break;
    if (j < _list.size())
    {
      DJV_DBG_PRN("duplicate");
      plugin.reset();
      continue;
    }
    
    // Add.
    
    plugin->init();
    add(plugin.release());
    _handle += handle.release();
  }
}

void Factory::add(Plugin * in)
{
  DJV_DBG_OBJ("Factory::add");
  DJV_DBG_PRN("in = " << in->name());
  
  _list += in;
}

void Factory::del()
{
  DJV_DBG_OBJ("Factory::del");

  for (uint i = 0; i < _list.size(); ++i)
  {
    _list[i]->del();
    delete _list[i];
  }
  _list.clear();

  for (uint i = 0; i < _handle.size(); ++i)
    delete _handle[i];
  _handle.clear();
}

const List<Plugin *> & Factory::list() const
{
  return _list;
}

List<Str> Factory::name_list() const
{
  List<Str> out(_list.size());
  for (uint i = 0; i < _list.size(); ++i)
    out[i] = _list[i]->name();
  return out;
}

}
}
