/*------------------------------------------------------------------------------
 lib/djv_image/cgl_offscreen_pbuffer.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#if defined(DJV_MACOSX)

//#define DJV_DBG

#include "cgl.h"
#include "cgl_offscreen.h"

namespace djv_image {
namespace gl_offscreen {

_Pbuffer::_Pbuffer() :
  _context(0),
  _pbuffer(0)
{
  DJV_DBG_OBJ("_Pbuffer::_Pbuffer");
}

_Pbuffer::~_Pbuffer()
{
  DJV_DBG_OBJ("_Pbuffer::~_Pbuffer");

  if (_pbuffer)
  {
    DJV_DBG_PRN("pbuffer");
    CGLDestroyPBuffer(_pbuffer);
  }
  
  if (_context)
  {
    DJV_DBG_PRN("context");
    CGLDestroyContext(_context);
  }
}

void _Pbuffer::init(const Info & in, const CGLPixelFormatObj format)
  throw (Error)
{
  DJV_DBG_OBJ("_Pbuffer::init");

  // Create context.

  CGLError error = CGLCreateContext(format, 0, &_context);
  if (error != kCGLNoError || ! _context)
    throw Error(gl::error_cgl,
      str::Format(gl::error_cgl_context_create).arg(error));

  error = CGLSetCurrentContext(_context);  
  if (error != kCGLNoError)
    throw Error(gl::error_cgl,
      str::Format(gl::error_cgl_context_bind).arg(error));

  // Create pbuffer.

  error = CGLCreatePBuffer(in.size.x, in.size.y, GL_TEXTURE_RECTANGLE_EXT,
    GL_RGBA, 0, &_pbuffer);
  if (error != kCGLNoError || ! _pbuffer)
    throw Error(gl::error_cgl,
      str::Format(error_cgl_pbuffer_create).arg(error));

  error = CGLSetPBuffer(_context, _pbuffer, 0, 0, 0);
  if (error != kCGLNoError)
    throw Error(gl::error_cgl, str::Format(error_cgl_pbuffer_set).arg(error));

  long w = 0, h = 0;
  unsigned long target = 0, internal_format = 0;
  long mipmap = 0;
  error = CGLDescribePBuffer(
    _pbuffer, &w, &h, &target, &internal_format, &mipmap);
  if (error != kCGLNoError)
    throw Error(gl::error_cgl,
      str::Format(error_cgl_pbuffer_describe).arg(error));

  _info.size = Vec2i(w, h);
  DJV_DBG_PRN("size = " << _info.size);
}

void _Pbuffer::current() throw (Error)
{
  DJV_DBG_OBJ("_Pbuffer::current");
  DJV_DBG_PRN("context = " << int(_context));

  CGLError error = CGLSetCurrentContext(_context);
  if (error != kCGLNoError)
    throw Error(gl::error_cgl,
      str::Format(gl::error_cgl_context_bind).arg(error));
}

_Cgl_Pbuffer_Factory::_Cgl_Pbuffer_Factory() :
  _format(0)
{
  DJV_DBG_OBJ("_Cgl_Pbuffer_Factory::_Cgl_Pbuffer_Factory");
}

_Cgl_Pbuffer_Factory::~_Cgl_Pbuffer_Factory()
{
  DJV_DBG_OBJ("_Cgl_Pbuffer_Factory::~_Cgl_Pbuffer_Factory");

  for (uint i = 0; i < _list.size(); ++i)
    delete _list[i];

  if (_format)
  {
    DJV_DBG_PRN("format");
    CGLDestroyPixelFormat(_format);
  }
}

void _Cgl_Pbuffer_Factory::init() throw (Error)
{
  DJV_DBG_OBJ("_Cgl_Pbuffer_Factory::init");

  // Pixel format.
  
  CGLPixelFormatAttribute attribs [] =
  {
    kCGLPFAPBuffer, (CGLPixelFormatAttribute)true,
    kCGLPFAColorFloat, (CGLPixelFormatAttribute)true,
    kCGLPFAColorSize, (CGLPixelFormatAttribute)32,
    (CGLPixelFormatAttribute)0
  };
  long npix = 0;
  CGLChoosePixelFormat(attribs, &_format, &npix);
  if (! _format)
  {
    // Fallback pixel format.

    attribs[2] = (CGLPixelFormatAttribute)0;
    CGLChoosePixelFormat(attribs, &_format, &npix);
  }
}

void _Cgl_Pbuffer_Factory::add(_Pbuffer * in)
{
  DJV_DBG_OBJ("_Cgl_Pbuffer_Factory::add");

  _list += in;
  if (int(_list.size()) > _list_max)
    delete _list.pop_front();
}

_Pbuffer * _Cgl_Pbuffer_Factory::get(const Info & in) throw (Error)
{
  DJV_DBG_OBJ("_Cgl_Pbuffer_Factory::get");
  DJV_DBG_PRN("in = " << in);
  
  _Pbuffer * out = 0;

  if (! _format)
    throw Error(gl::error_cgl, str::Format(gl::error_cgl_pixel_format));

  for (List<_Pbuffer *>::iterator i = _list.begin(); i != _list.end(); ++i)
    if (in == (*i)->info())
    {
      DJV_DBG_PRN("cached");
      out = *i;
      _list.erase(i);
      break;
    }
  
  if (! out)
  {
    out = new _Pbuffer;
    out->init(in, _format);
  }
  return out;
}

class Global::_Pbuffer::Context_Restore
{
public:

  Context_Restore() :
    _context(CGLGetCurrentContext())
  {}
  
  ~Context_Restore()
  {
    CGLSetCurrentContext(_context);
  }
  
private:

  CGLContextObj _context;
};

Global::_Pbuffer::_Pbuffer(_Pbuffer_Factory * factory) :
  _factory(factory),
  _p(0)
{
  DJV_DBG_OBJ("Global::_Pbuffer::_Pbuffer");
}

Global::_Pbuffer::~_Pbuffer()
{
  DJV_DBG_OBJ("Global::_Pbuffer::~_Pbuffer");
  
  if (_p) _factory->add(_p);
}

void Global::_Pbuffer::init(const Info & in) throw (Error)
{
  DJV_DBG_OBJ("Global::_Pbuffer::init");
  DJV_DBG_PRN("in = " << in);

  Context_Restore context_restore;
  
  if (_p)
    if (in != _p->info())
    {
      _factory->add(_p);
      _p = 0;
    }

  if (! _p)
    _p = _factory->get(in);
}

const Info & Global::_Pbuffer::info() const
{
  return _p->info();
}

void Global::_Pbuffer::bind()
{
  DJV_DBG_OBJ("Global::_Pbuffer::bind");
  
  _context_restore = std::auto_ptr<Context_Restore>(new Context_Restore);
  
  _p->current();
}

void Global::_Pbuffer::unbind()
{
  DJV_DBG_OBJ("Global::_Pbuffer::unbind");
  
  _context_restore.reset();
}

}
}

#endif
