/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_program_etc.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "gl_data_program.h"

namespace djv_image {
namespace gl_data {
namespace program {

namespace {

const Str
  error_name = "OpenGL",
  error_texture = "Texture Create",
  error_shader_create = "Shader Create",
  error_shader_compile = "Shader Compile:\n%%",
  error_shader_program = "Shader Program",
  error_shader_link = "Shader Link:\n%%";

}

Lut::Lut() :
  _size(0),
  _id(0)
{}

Lut::~Lut()
{
  del();
}

void Lut::init(const djv_image::Info & in) throw (Error)
{
  if (in == _info) return;
  
  DJV_DBG_OBJ("Lut::init");
  DJV_DBG_PRN("in = " << in);
  
  _info = in;
  _size = math::to_pow2(_info.size.x);
  DJV_DBG_PRN("size = " << _size);

  del();

  DJV_DBG_GL(glGenTextures(1, &_id));
  if (! _id)
    throw Error(error_name, error_texture);

  DJV_DBG_GL(glBindTexture(GL_TEXTURE_1D, _id));
  DJV_DBG_GL(glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP));
  DJV_DBG_GL(glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST));
  DJV_DBG_GL(glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST));
  
  DJV_DBG_GL(glTexImage1D(
    GL_TEXTURE_1D, 0, GL_RGBA,
    _size, 0,
    gl::format(_info.pixel, _info.bgr),
    gl::type(_info.pixel),
    0));
}

void Lut::init(const djv_image::Data & in) throw (Error)
{
  DJV_DBG_OBJ("Lut::init");
  DJV_DBG_PRN("in = " << in);
  
  init(in.info());
  bind();
  copy(in);
}

void Lut::del()
{
  if (_id)
  {
    glDeleteTextures(1, &_id);
    _id = 0;
  }
}

void Lut::copy(const djv_image::Data & in)
{
  DJV_DBG_OBJ("Lut::copy");
  DJV_DBG_PRN("in = " << in);
  
  const djv_image::Info & info = in.info();

  djv_image::gl_data::state_unpack(in.info());
  
  DJV_DBG_GL(glTexSubImage1D(
    GL_TEXTURE_1D, 0,
    0, info.size.x,
    gl::format(info.pixel, info.bgr),
    gl::type(info.pixel),
    in.data()));
}

void Lut::bind()
{
  DJV_DBG_OBJ("Lut::bind");
  
  DJV_DBG_GL(glBindTexture(GL_TEXTURE_1D, _id));
}

const djv_image::Info & Lut::info() const { return _info; }
GLuint Lut::id() const { return _id; }

Texture::Texture() :
  _min(GL_LINEAR), _mag(GL_LINEAR),
  _id(0)
{}

Texture::~Texture()
{
  del();
}

void Texture::init(const djv_image::Info & in, GLenum min, GLenum mag)
  throw (Error)
{
  if (in == _info && min == _min && mag == _mag) return;
  
  DJV_DBG_OBJ("Texture::init");
  DJV_DBG_PRN("in = " << in);
  
  _info = in;
  _min = min;
  _mag = mag;
  DJV_DBG_PRN("min = " << _min);
  DJV_DBG_PRN("mag = " << _mag);

  del();

  DJV_DBG_GL(glGenTextures(1, &_id));
  if (! _id)
    throw Error(error_name, error_texture);

  const GLenum target = gl::texture();
  DJV_DBG_GL(glBindTexture(target, _id));
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP));
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP));
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_MIN_FILTER, _min));
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_MAG_FILTER, _mag));
  
  // XXX
  
  GLenum format = GL_RGBA;
  if (
    djv_image::pixel::F16 == djv_image::pixel::type(_info.pixel) ||
    djv_image::pixel::F32 == djv_image::pixel::type(_info.pixel)
  ) {
    if (GLEW_NV_float_buffer)
      format = GL_FLOAT_RGBA_NV;
    else if(GLEW_ARB_texture_float)
      format = GL_RGBA32F_ARB;
  }
  
  DJV_DBG_GL(glTexImage2D(
    target, 0, format,
    _info.size.x, _info.size.y, 0,
    djv_image::gl::format(_info.pixel, _info.bgr),
    djv_image::gl::type(_info.pixel),
    0));
}

void Texture::init(const djv_image::Data & in, GLenum min, GLenum mag)
  throw (Error)
{
  DJV_DBG_OBJ("Texture::init");
  DJV_DBG_PRN("in = " << in);
  
  init(in.info(), min, mag);
  bind();
  copy(in);
}

void Texture::del()
{
  if (_id)
  {
    glDeleteTextures(1, &_id);
    _id = 0;
  }
}

void Texture::copy(const djv_image::Data & in)
{
  DJV_DBG_OBJ("Texture::copy");
  DJV_DBG_PRN("in = " << in);
  
  const djv_image::Info & info = in.info();

  djv_image::gl_data::state_unpack(in.info());
  
  DJV_DBG_GL(glTexSubImage2D(
    gl::texture(), 0,
    0, 0, info.size.x, info.size.y,
    djv_image::gl::format(info.pixel, info.bgr),
    djv_image::gl::type(info.pixel),
    in.data()));
}

void Texture::copy(const Vec2i & in)
{
  DJV_DBG_OBJ("Texture::copy");
  DJV_DBG_PRN("in = " << in);

  DJV_DBG_GL(glCopyTexSubImage2D(
    gl::texture(), 0,
    0, 0, 0, 0, in.x, in.y));
}

void Texture::bind()
{
  DJV_DBG_OBJ("Texture::bind");
  
  DJV_DBG_GL(glBindTexture(gl::texture(), _id));
}

const djv_image::Info & Texture::info() const { return _info; }
GLenum Texture::min() const { return _min; }
GLenum Texture::mag() const { return _mag; }
GLuint Texture::id() const { return _id; }

Shader::Shader() :
  _id(0),
  _program(0)
{}

Shader::~Shader()
{
  del();
}

void Shader::init(const Str & in) throw (Error)
{
  DJV_DBG_OBJ("Shader::init");
  DJV_DBG_PRN("in = " << in);
  
  GLint error = 0;
  
  del();
  
  DJV_DBG_GL(_id = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB));
  if (! _id)
    throw Error(error_name, error_shader_create);
  
  const char * src_list [] = { in.c_str() };
  DJV_DBG_GL(glShaderSourceARB(_id, 1, src_list, 0));
  DJV_DBG_GL(glCompileShaderARB(_id));
  char log [str::cstr_size] = "";
  glGetInfoLogARB(_id, str::cstr_size, 0, log);
  DJV_DBG_PRN("log = " << log);
  glGetObjectParameterivARB(_id, GL_OBJECT_COMPILE_STATUS_ARB, &error);
  if (! error)
    throw Error(error_name, Str(str::Format(error_shader_compile).arg(log)));

  DJV_DBG_GL(_program = glCreateProgramObjectARB());
  if (! _program)
    throw Error(error_name, error_shader_program);

  DJV_DBG_GL(glAttachObjectARB(_program, _id));
  DJV_DBG_GL(glLinkProgramARB(_program));
  glGetInfoLogARB(_program, str::cstr_size, 0, log);
  DJV_DBG_PRN("log = " << log);
  glGetObjectParameterivARB(_id, GL_OBJECT_LINK_STATUS_ARB, &error);
  if (! error)
    throw Error(error_name, Str(str::Format(error_shader_link).arg(log)));
}

void Shader::bind()
{
  DJV_DBG_OBJ("Shader::bind");
  
  DJV_DBG_GL(glUseProgramObjectARB(_program));
}

void Shader::unbind()
{
  glUseProgramObjectARB(0);
}

void Shader::del()
{
  if (_id)
  {
    glDeleteObjectARB(_id);
    _id = 0;
  }
  if (_program)
  {
    glDeleteObjectARB(_program);
    _program = 0;
  }
}

GLhandleARB Shader::program() const { return _program; }

Program::State_Default::State_Default() :
  _init(false)
{}

Program::State_Default::State_Default(const Option & in) :
  _profile(in.profile.type),
  _channel(in.channel),
  _display(in.display),
  _filter(in.filter._default),
  _clamp(in.clamp),
  _init(true)
{}
    
bool Program::State_Default::operator != (const State_Default & in) const
{
  return
    in._profile != _profile ||
    in._channel != _channel ||
    in._display.lut.info() != _display.lut.info() ||
    in._display.color != _display.color ||
    in._display.levels != _display.levels ||
    in._display.soft_clip != _display.soft_clip ||
    in._filter != _filter ||
    in._clamp != _clamp ||
    ! _init;
}

Program::State_Custom::State_Custom() :
  _init(false)
{}

Program::State_Custom::State_Custom(const Vec2i & size, int proxy_scale,
  const Option & option) :
  _size(size),
  _proxy_scale(proxy_scale),
  _profile(option.profile.type),
  _channel(option.channel),
  _display(option.display),
  _filter(option.filter.custom),
  _clamp(option.clamp),
  _init(true)
{}
    
bool Program::State_Custom::operator != (const State_Custom & in) const
{
  return
    in._size != _size ||
    in._proxy_scale != _proxy_scale ||
    in._profile != _profile ||
    in._channel != _channel ||
    in._display.lut.info() != _display.lut.info() ||
    in._display.color != _display.color ||
    in._display.levels != _display.levels ||
    in._display.soft_clip != _display.soft_clip ||
    in._filter != _filter ||
    in._clamp != _clamp ||
    ! _init;
}

void quad(const Vec2i & size, const Vec2b & mirror, int proxy_scale)
{
  double u [] = { 0, 0 }, v [] = { 0, 0 };
  u[! mirror.x] = size.x;
  v[! mirror.y] = size.y;
  if (GL_TEXTURE_2D == gl::texture())
  {
    u[! mirror.x] /= double(size.x);
    v[! mirror.y] /= double(size.y);
  }
  const Vec2f uv [] =
  {
    Vec2f(u[0], v[0]),
    Vec2f(u[0], v[1]),
    Vec2f(u[1], v[1]),
    Vec2f(u[1], v[0])
  };
      
  glBegin(GL_QUADS);
  gl::box(Box2i(size * proxy_scale), uv);
  glEnd();
}

}
}
}
