/*------------------------------------------------------------------------------
 lib/djv_image/gl_offscreen_fbo.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "gl_offscreen.h"
#include <djv_image/gl_data.h>

namespace djv_image {
namespace gl_offscreen {

namespace {

Str error(GLenum in)
{
  switch (in)
  {
    case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
      return "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT";
    case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
      return "GL_FRAMEBUFFER_UNSUPPORTED_EXT";
    case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
      return "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT";
    case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
      return "GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT";
    case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
      return "GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT";
    case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
      return "GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT";
    case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
      return "GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT";
  }
  return "";
}

}

_Fbo::_Fbo(const Info & in) :
  _info(in),
  _id(0),
  _restore(0),
  _texture(0)
{
  if (! GLEW_EXT_framebuffer_object) return;
  
  DJV_DBG_OBJ("_Fbo::_Fbo");
  DJV_DBG_PRN("in = " << in);

  DJV_DBG_GL(glGenTextures(1, &_texture));

  const GLenum target = gl::texture();
  DJV_DBG_GL(glBindTexture(target, _texture));
  if (! _texture) return;
  
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP));
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP));
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST));
  DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST));

  // XXX
  
  GLenum format = GL_RGBA;
  if (GLEW_NV_float_buffer)
    format = GL_FLOAT_RGBA_NV;
  else if(GLEW_ARB_texture_float)
    format = GL_RGBA32F_ARB;
  
  DJV_DBG_GL(glTexImage2D(
    target, 0, format,
    _info.size.x, _info.size.y, 0,
    djv_image::gl::format(_info.pixel, _info.bgr),
    djv_image::gl::type(_info.pixel),
    0));
  
  DJV_DBG_GL(glGenFramebuffersEXT(1, &_id));
  if (! _id) return;
  
  DJV_DBG_GL(glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id));
  DJV_DBG_GL(glFramebufferTexture2DEXT(
    GL_FRAMEBUFFER_EXT,
    GL_COLOR_ATTACHMENT0_EXT,
    target, _texture, 0));
  DJV_DBG_PRN("error = " <<
    error(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT)));
}

_Fbo::~_Fbo()
{
  if (! GLEW_EXT_framebuffer_object) return;
  
  DJV_DBG_OBJ("_Fbo::~_Fbo");
  
  if (_id)
  {
    DJV_DBG_GL(glDeleteFramebuffersEXT(1, &_id));
  }
  if (_texture)
  {
    DJV_DBG_GL(glDeleteTextures(1, &_texture));
  }
}

namespace {

GLuint _current = 0; // XXX

}

void _Fbo::bind()
{
  if (! GLEW_EXT_framebuffer_object || ! _id) return;
  
  DJV_DBG_OBJ("_Fbo::bind");
  DJV_DBG_PRN("current = " << int(_current));
  
  DJV_DBG_GL(glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id));
  _restore = _current;
  _current = _id;
}

void _Fbo::unbind()
{
  if (! GLEW_EXT_framebuffer_object || ! _id) return;
  
  DJV_DBG_OBJ("_Fbo::unbind");
  
  DJV_DBG_GL(glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _restore));
  _current = _restore;
  DJV_DBG_PRN("current = " << int(_current));
}

void _Fbo::texture_bind()
{
  if (! _texture) return;
  
  DJV_DBG_OBJ("_Fbo::texture_bind");

  DJV_DBG_GL(glBindTexture(gl::texture(), _texture));
}

_Fbo_Factory::_Fbo_Factory()
{
  DJV_DBG_OBJ("_Fbo_Factory::_Fbo_Factory");
}

_Fbo_Factory::~_Fbo_Factory()
{
  DJV_DBG_OBJ("_Fbo_Factory::~_Fbo_Factory");
  
  for (uint i = 0; i < _list.size(); ++i)
    delete _list[i];
}

void _Fbo_Factory::init() throw (Error)
{
  DJV_DBG_OBJ("_Fbo_Factory::init");
  DJV_DBG_PRN("valid = " << GLEW_EXT_framebuffer_object);
}

void _Fbo_Factory::add(_Fbo * in)
{
  DJV_DBG_OBJ("_Fbo_Factory::add");
  
  if (0)
  {
    delete in;
  }
  else
  {
    _list += in;
    if (int(_list.size()) > _list_max)
      delete _list.pop_front();
  }
}

_Fbo * _Fbo_Factory::get(const Info & in) throw (Error)
{
  DJV_DBG_OBJ("_Fbo_Factory::get");
  DJV_DBG_PRN("in = " << in);
  
  _Fbo * out = 0;

  for (List<_Fbo *>::iterator i = _list.begin(); i != _list.end(); ++i)
    if (in == (*i)->info())
    {
      DJV_DBG_PRN("cached = " << (*i)->info().size);
      out = *i;
      _list.erase(i);
      break;
    }
  
  if (! out) out = new _Fbo(in);

  return out;
}

Global::_Fbo::_Fbo(_Fbo_Factory * factory) :
  _factory(factory),
  _p(0)
{
  DJV_DBG_OBJ("Global::_Fbo::_Fbo");
}

Global::_Fbo::~_Fbo()
{
  DJV_DBG_OBJ("Global::_Fbo::~_Fbo");
  
  if (_p) _factory->add(_p);
}

void Global::_Fbo::init(const Info & in) throw (Error)
{
  DJV_DBG_OBJ("Global::_Fbo::init");
  DJV_DBG_PRN("in = " << in);
  
  if (_p)
    if (in != _p->info())
    {
      _factory->add(_p);
      _p = 0;
    }

  if (! _p)
    _p = _factory->get(in);
}

void Global::_Fbo::bind()
{
  DJV_DBG_OBJ("Global::_Fbo::bind");
  
  _p->bind();
}

void Global::_Fbo::texture_bind()
{
  DJV_DBG_OBJ("Global::_Fbo::texture_bind");
  
  _p->texture_bind();
}

void Global::_Fbo::unbind()
{
  DJV_DBG_OBJ("Global::_Fbo::unbind");
  
  _p->unbind();
}

const Info & Global::_Fbo::info() const { return _p->info(); }

}
}
