/*------------------------------------------------------------------------------
 lib/djv_image/io_ppm_save.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_image/io_ppm.h>

namespace djv_image {
namespace io_ppm {

Save::Save()
{
  _io.endian(mem::endian() != mem::MSB);
}

io::Plugin * Save::copy() const
{
  Save * out = new Save;
  out->_option = _option;
  return out;
}

const Str & Save::name() const { return io_ppm::name; }

const List<Str> & Save::extension_list() const
  { return io_ppm::extension_list; }

bool Save::option(const Str & in, Str * data)
{
try
{
  if (str::compare(in, option_list()[SAVE_TYPE])) *data >> _option.type;
  else if (str::compare(in, option_list()[SAVE_MODE])) *data >> _option.mode;
}
catch (Str) { return false; }
  return true;
}

Str Save::option(const Str & in) const
{
  Str out;
  if (str::compare(in, option_list()[SAVE_TYPE])) out << _option.type;
  else if (str::compare(in, option_list()[SAVE_MODE])) out << _option.mode;
  return out;
}

const List<Str> & Save::option_list() const
{
  return label_save_option();
}

void Save::open(const File & in, const Info & info, const Tag &,
  const Time & time) throw (Error)
{
  DJV_DBG_OBJ("Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  if (time.list.size() > 1) _file.type = file::SEQ;
  
  _info = Info();
  _info.size = info.size;
  _info.mirror.y = true;
  switch (_option.type)
  {
    case TYPE_AUTO:
    {
      pixel::FORMAT format = pixel::format(info.pixel);
      switch (format)
      {
        case pixel::LA: format = pixel::L; break;
        case pixel::RGBA: format = pixel::RGB; break;
        default: break;
      }
      pixel::TYPE type = pixel::type(info.pixel);
      switch (type)
      {
        case pixel::U10:
        case pixel::F16: 
        case pixel::F32: type = pixel::U16; break;
        default: break;
      }
      _info.pixel = pixel::pixel(format, type);
      _bit_depth = pixel::bit_depth(_info.pixel);
    } break;
    case U1: _info.pixel = pixel::L_U8; _bit_depth = 1; break;
  }
  _info.endian = mem::MSB;
  DJV_DBG_PRN("info = " << _info);

  _image.set(_info);
}

void Save::_open(const File & in) throw (Error)
{
  DJV_DBG_OBJ("Save::_open");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _io.open(in, djv_base::io::WRITE);
  
  // Header.
  
  int ppm_type = 0;
  if (1 == _bit_depth)
  {
    ppm_type = ASCII == _option.mode ? 1 : 4;
  }
  else
  {
    ppm_type = ASCII == _option.mode ? 2 : 5;
    if (3 == _image.channels()) ++ppm_type;
  }
  char magic [] = "P \n";
  magic[1] = '0' + ppm_type;
  _io.set(magic, 3);
  
  char tmp [str::cstr_size];
  int size = SNPRINTF(tmp, str::cstr_size, "%d %d\n", _image.w(), _image.h());
  _io.set(tmp, size);
  
  if (_bit_depth != 1)
  {
    const int max_value = (8 == _bit_depth) ? pixel::u8_max : pixel::u16_max;
    size = SNPRINTF(tmp, str::cstr_size, "%d\n", max_value);
    _io.set(tmp, size);
  }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ("Save::set");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("type = " << _option.type);
  DJV_DBG_PRN("mode = " << _option.mode);
  
  // Open.

  _open(_file.get(frame));
  
  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl_data::copy(in, &_image);
    p = &_image;
  }
  
  if (BINARY == _option.mode && _bit_depth != 1)
  {
    _io.set(p->data(), p->bytes_data());
  }
  else
  {
    const int w = p->w(), h = p->h();
    const int channels = pixel::channels(p->info().pixel);
    const size_t bytes_scanline = io_ppm::bytes_scanline(w, channels,
      _bit_depth, _option.mode);
    mem::Buffer<uint8_t> scanline(bytes_scanline);
    DJV_DBG_PRN("scanline = " << int(bytes_scanline));
  
    for (int y = 0; y < h; ++y)
      if (BINARY == _option.mode && 1 == _bit_depth)
      {
        const uint8_t * in_p = p->data(0, y);
        uint8_t * out_p = scanline();
        for (int i = 0; i < w; ++i)
        {
          const int tmp = in_p[i];
          const int off = i % 8;
          const int j = i / 8;
          if (0 == off) out_p[j] = 0;
          out_p[j] |= ((! tmp) & 1) << (7 - off);
        }
        _io.set(scanline(), bytes_scanline);
      }
      else
      {
        const size_t size = ascii_save(p->data(0, y), scanline(), w * channels,
          _bit_depth);
        _io.set(scanline(), size);
      }
  }
  
  _io.close();
}

}
}
