/*------------------------------------------------------------------------------
 lib/djv_image/io_rla_load.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_image/io_rla.h>

namespace djv_image {
namespace io_rla {

Load::Load()
{
  _io.endian(mem::endian() != mem::MSB);
}

io::Plugin * Load::copy() const { return new Load; }

const Str & Load::name() const { return io_rla::name; }

const List<Str> & Load::extension_list() const
  { return io_rla::extension_list; }

void Load::open(const File & in, Info * info, Tag * tag, Time * time)
  throw (Error)
{
  io::Load::open(in, info, tag, time);
}

void Load::open(const File & in, List<Info> * info, Tag * tag, Time * time)
  throw (Error)
{
  DJV_DBG_OBJ("Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start()), info);

  if (file::SEQ == _file.type)
    time->list = _file.seq.list;
}

namespace {

struct Header
{
  int16_t dimensions [4]; // Left, right, bottom, top.
  int16_t active [4];
  int16_t frame;
  int16_t color_channel_type;
  int16_t color_channels, matte_channels, aux_channels;
  int16_t version;
  char gamma [16];
  char chroma [3][24], whitepoint [24];
  int32_t job;
  char filename [128];
  char description [128];
  char progam [64];
  char machine [32];
  char user [32];
  char date [20];
  char aspect [24], aspect_ratio [8];
  char color_format [32];
  int16_t field;
  char render_time [12];
  char filter [32];
  int16_t color_bit_depth;
  int16_t matte_channel_type;
  int16_t matte_bit_depth;
  int16_t aux_channel_type;
  int16_t aux_bit_depth;
  char aux_format [32];
  char pad [36];
  int32_t offset;
};

void endian(Header * in)
{
  mem::endian(&in->dimensions, 4, 2);
  mem::endian(&in->active, 4, 2);
  mem::endian(&in->frame, 1, 2);
  mem::endian(&in->color_channel_type, 1, 2);
  mem::endian(&in->color_channels, 1, 2);
  mem::endian(&in->matte_channels, 1, 2);
  mem::endian(&in->aux_channels, 1, 2);
  mem::endian(&in->version, 1, 2);
  mem::endian(&in->job, 1, 4);
  mem::endian(&in->field, 1, 2);
  mem::endian(&in->color_bit_depth, 1, 2);
  mem::endian(&in->matte_channel_type, 1, 2);
  mem::endian(&in->matte_bit_depth, 1, 2);
  mem::endian(&in->aux_channel_type, 1, 2);
  mem::endian(&in->aux_bit_depth, 1, 2);
  mem::endian(&in->offset, 1, 4);
}

void dbg(const Header & in)
{
  DJV_DBG_OBJ("dbg(Header)");
  DJV_DBG_PRN("dimensions = " << in.dimensions[0] << " " << in.dimensions[1] <<
    " " << in.dimensions[2] << " " << in.dimensions[3]);
  DJV_DBG_PRN("active = " << in.active[0] << " " << in.active[1] << " " <<
    in.active[2] << " " << in.active[3]);
  DJV_DBG_PRN("frame = " << in.frame);
  DJV_DBG_PRN("color channel type = " << in.color_channel_type);
  DJV_DBG_PRN("color channels = " << in.color_channels);
  DJV_DBG_PRN("matte channels = " << in.matte_channels);
  DJV_DBG_PRN("aux channels = " << in.aux_channels);
  DJV_DBG_PRN("version = " << in.version);
  DJV_DBG_PRN("gamma = " << in.gamma);
  DJV_DBG_PRN("chroma = " << in.chroma[0] << " " << in.chroma[1] << " " <<
    in.chroma[2]);
  DJV_DBG_PRN("whitepoint = " << in.whitepoint);
  DJV_DBG_PRN("job = " << in.job);
  DJV_DBG_PRN("filename = " << in.filename);
  DJV_DBG_PRN("description = " << in.description);
  DJV_DBG_PRN("progam = " << in.progam);
  DJV_DBG_PRN("machine = " << in.machine);
  DJV_DBG_PRN("user = " << in.user);
  DJV_DBG_PRN("date = " << in.date);
  DJV_DBG_PRN("aspect = " << in.aspect);
  DJV_DBG_PRN("aspect ratio = " << in.aspect_ratio);
  DJV_DBG_PRN("color format = " << in.color_format);
  DJV_DBG_PRN("field = " << in.field);
  DJV_DBG_PRN("render time = " << in.render_time);
  DJV_DBG_PRN("filter = " << in.filter);
  DJV_DBG_PRN("color bit depth = " << in.color_bit_depth);
  DJV_DBG_PRN("matte channel type = " << in.matte_channel_type);
  DJV_DBG_PRN("matte bit depth = " << in.matte_bit_depth);
  DJV_DBG_PRN("aux channel type = " << in.aux_channel_type);
  DJV_DBG_PRN("aux bit depth = " << in.aux_bit_depth);
  DJV_DBG_PRN("aux format = " << in.aux_format);
  DJV_DBG_PRN("offset = " << in.offset);
}

}

void Load::_open(const File & in, List<Info> * info) throw (Error)
{
  DJV_DBG_OBJ("Load::_open");
  DJV_DBG_PRN("in = " << in);
  
  // Open.
  
  _io.open(in, djv_base::io::READ);
  
  // Header.
  
  Header header;
  DJV_DBG_PRN("header size = " << int(sizeof(Header)));
  _io.get(&header, sizeof(Header));
  if (_io.endian()) endian(&header);
  dbg(header);
  
  const int w = header.active[1] - header.active[0] + 1;
  const int h = header.active[3] - header.active[2] + 1;
  
  // Scanline table.
  
  _rle_offset.size(h);
  _io.get_32(_rle_offset(), h);

  // Information.
  
  const Vec2i size(w, h);
  PIXEL pixel(PIXEL(0));

  if (header.matte_channels > 1)
    throw Error(name(), io::error_unsupported);
  
  if (header.matte_channel_type != header.color_channel_type)
    throw Error(name(), io::error_unsupported);

  if (header.matte_bit_depth != header.color_bit_depth)
    throw Error(name(), io::error_unsupported);
  
  if (! pixel::pixel_(
    header.color_channels + header.matte_channels,
    header.color_bit_depth,
    3 == header.color_channel_type,
    &pixel
  ))
    throw Error(name(), io::error_unsupported);

  if (header.field)
    throw Error(name(), io::error_unsupported);
  
  info->push_back(Info(in, size, pixel));
}

const Image * Load::get(int frame, int layer, data::PROXY proxy) throw (Error)
{
  DJV_DBG_OBJ("Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("layer = " << layer);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  List<Info> info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start()), &info);

  if (layer < 0 || layer >= int(info.size()))
    throw Error(name(), djv_base::io::error_read);

  // Read.

  _io.readahead();
  
  Data * p = proxy ? &_tmp : &_image;
  p->set(info[layer]);

  const int w = info[layer].size.x, h = info[layer].size.y;
  const int channels = pixel::channels(info[layer].pixel);
  const int bytes = pixel::channel_bytes(info[layer].pixel);
  DJV_DBG_PRN("channels = " << channels);
  DJV_DBG_PRN("bytes = " << bytes);
  
  uint8_t * data_p = p->data();
  for (int y = 0; y < h; ++y, data_p += w * channels * bytes)
  {    
    _io.position(_rle_offset()[y]);
    
    for (int c = 0; c < channels; ++c)
      if (pixel::F32 == pixel::type(info[layer].pixel))
        float_load(&_io, data_p + c * bytes, w, channels);
      else
        rle_load(&_io, data_p + c * bytes, w, channels, bytes);
  }

  if (proxy)
  {
    info[layer].size = data::proxy_scale(info[layer].size, proxy);
    info[layer].proxy = proxy;
    _image.set(info[layer]);
    data::proxy_scale(_tmp, &_image, proxy);
  }

  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}
