/*------------------------------------------------------------------------------
 lib/djv_ui/app.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_ui/app.h>
#include <FL/Fl.H>

namespace djv_ui {
namespace app {

App::App(const Str & name, int & argc, char ** argv) throw (Error) :
  djv_image::app::App(name, argc, argv),
  _valid(false),
  _prefs(new prefs::Prefs(name)),
  _prefs_global(new prefs::Prefs("global")),
  _base_prefs(new base::Prefs),
  _style_prefs(new style::Prefs),
  _image_prefs(new image::Prefs),
  _file_browser_prefs(new file_browser::Prefs)
{
  DJV_DBG_OBJ("App::App");
}

App::~App()
{
  DJV_DBG_OBJ("App::~App");

  /*// Debug.
  
  DJV_DBG_PRN("widgets = " << _widget_list.size());
  for (uint i = 0; i < _widget_list.size(); ++i)
    DJV_DBG_PRN("  " << _widget_list[i]->dbg_name());
  DJV_DBG_PRN("windows = " << _window_list.size());*/
}

int App::exec()
{
  DJV_DBG_OBJ("App::exec");
  DJV_DBG_PRN("valid = " << valid());
  DJV_DBG_PRN("widgets = " << int(_widget_list.size()));
  DJV_DBG_PRN("windows = " << int(_window_list.size()));
  
  if (djv_image::app::App::exit() != djv_base::app::EXIT_DEFAULT) exit();
  
  Fl::run();
  
  return djv_image::app::App::exec();
}

void App::valid(bool in)
{
  DJV_DBG_OBJ("App::valid");
  DJV_DBG_PRN("in = " << in);
  
  _valid = in;
  
  if (_valid)
  {
    _base = std::auto_ptr<base::Global>(new base::Global);
    _style = std::auto_ptr<style::Style>(new style::Style);
    _icon = std::auto_ptr<icon::Factory>(new icon::Factory);
    _image_io = std::auto_ptr<image_io::Global>(new image_io::Global);
    _dialog = std::auto_ptr<dialog::Global>(new dialog::Global);
    _prefs_dialog.p = std::auto_ptr<prefs_dialog::Dialog>(
      new prefs_dialog::Dialog);
    _messages_dialog.p = std::auto_ptr<messages_dialog::Dialog>(
      new messages_dialog::Dialog);
    _file_browser.p = std::auto_ptr<file_browser::Dialog>(
      new file_browser::Dialog);
    _app_dialog = std::auto_ptr<app_dialog::Global>(new app_dialog::Global);
  }
}

bool App::valid() const { return _valid; }

void App::exit(djv_base::app::EXIT in)
{
  DJV_DBG_OBJ("App::exit");
  DJV_DBG_PRN("widgets = " << int(_widget_list.size()));
  DJV_DBG_PRN("windows = " << int(_window_list.size()));
  
  // Close all windows.
  
  for (uint i = 0; i < _window_list.size(); ++i)
    _window_list[i]->hide();
  
  if (in != djv_base::app::EXIT_DEFAULT) djv_image::app::App::exit(in);
}

Str App::info() const
{
  return djv_image::app::App::info();
  
  // XXX
  
  /*return str::Format(label_info).
    arg(djv_image::app::App::info()).
    arg(str::join(image_io::global()->load()->name_list(), ", ")).
    arg(str::join(image_io::global()->save()->name_list(), ", "));*/
}

void App::message(const Str & in) const
{
  if (valid()) messages_dialog::message(in);
  else djv_image::app::App::message(in);
}

void App::error(const Error & in, const Str & message) const
{
  if (valid()) messages_dialog::error(in, message);
  else djv_image::app::App::error(in, message);
}

void App::widget_add(widget::Base * in)
{
  //DJV_DBG_OBJ("App::widget_add");
  
  _widget_list += in;
}

void App::widget_del(widget::Base * in)
{
  //DJV_DBG_OBJ("App::widget_del");
  
  List<widget::Base *>::iterator i = std::find(
    _widget_list.begin(), _widget_list.end(), in);
  if (i != _widget_list.end()) _widget_list.erase(i);
}

void App::window_add(window::Widget * in)
{
  //DJV_DBG_OBJ("App::window_add");
  
  _window_list += in;
}

void App::window_del(window::Widget * in)
{
  //DJV_DBG_OBJ("App::window_del");
  
  List<window::Widget *>::iterator i = std::find(
    _window_list.begin(), _window_list.end(), in);
  if (i != _window_list.end()) _window_list.erase(i);
}

void App::window_redraw()
{
  DJV_DBG_OBJ("App::window_redraw");
  
  for (uint i = 0; i < _window_list.size(); ++i)
    _window_list[i]->redraw();
}

void App::window_update()
{
  DJV_DBG_OBJ("App::window_update");
  
  for (uint i = 0; i < _window_list.size(); ++i)
  {
    _window_list[i]->dirty();
    _window_list[i]->update();
  }
}

prefs::Prefs * App::prefs() const { return _prefs.get(); }
prefs::Prefs * App::prefs_global() const { return _prefs_global.get(); }
base::Global * App::base() const { return _base.get(); }
base::Prefs * App::base_prefs() const { return _base_prefs.get(); }
style::Style * App::style() const { return _style.get(); }
style::Prefs * App::style_prefs() const { return _style_prefs.get(); }
icon::Factory * App::icon() const { return _icon.get(); }
image::Prefs * App::image_prefs() const { return _image_prefs.get(); }
image_io::Global * App::image_io() const { return _image_io.get(); }
dialog::Global * App::dialog() const { return _dialog.get(); }
prefs_dialog::Dialog * App::prefs_dialog() const { return _prefs_dialog.get(); }
messages_dialog::Dialog * App::messages_dialog() const
  { return _messages_dialog.get(); }
file_browser::Dialog * App::file_browser() const { return _file_browser.get(); }
file_browser::Prefs * App::file_browser_prefs() const
  { return _file_browser_prefs.get(); }
app_dialog::Global * App::app_dialog() const { return _app_dialog.get(); }

const Str label_info =
"%%"
"\n"
"UI Image I/O:\n"
"* Load: %%\n"
"* Save: %%\n";

App * global() { return static_cast<App *>(djv_image::app::global()); }

}

namespace prefs {
Prefs * prefs() { return app::global()->prefs(); }
Prefs * prefs_global() { return app::global()->prefs_global(); }
}
namespace base {
Global * global() { return app::global()->base(); }
Prefs * prefs() { return app::global()->base_prefs(); }
}
namespace style {
Style * global() { return app::global()->style(); }
Prefs * prefs() { return app::global()->style_prefs(); }
}
namespace icon {
Factory * global() { return app::global()->icon(); }
}
namespace image {
Prefs * prefs() { return app::global()->image_prefs(); }
}
namespace image_io {
Global * global() { return app::global()->image_io(); }
}
namespace dialog {
Global * global() { return app::global()->dialog(); }
}
namespace prefs_dialog {
Dialog * global() { return app::global()->prefs_dialog(); }
}
namespace messages_dialog {
Dialog * global() { return app::global()->messages_dialog(); }
}
namespace file_browser {
Dialog * global() { return app::global()->file_browser(); }
Prefs * prefs() { return app::global()->file_browser_prefs(); }
}
namespace app_dialog {
Global * global() { return app::global()->app_dialog(); }
}

}

