/*------------------------------------------------------------------------------
 lib/djv_ui/menu.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include <djv_ui/menu.h>
#include <djv_ui/style.h>
#include <FL/Fl.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/fl_draw.H>

namespace djv_ui {
namespace menu {

Item::Item() :
  shortcut(0),
  callback(0),
  data(0),
  flags(0)
{}

Item::Item(
  const Str & label,
  int shortcut,
  Fl_Callback * callback,
  void * data,
  int flags
) :
  label(label),
  shortcut(shortcut),
  callback(callback),
  data(data),
  flags(flags)
{
  // Convert label to a c-string, chopping it if necessary.
  
  const int size = math::min(int(label.size()), item_label_max_size);
  cstr.size(size + 1);
  mem::copy(label.c_str(), cstr(), size);
  cstr()[size] = 0;

  if (int(label.size()) > item_label_max_size)
  {
    DJV_ASSERT(item_label_max_size >= 3);
    
    cstr()[size - 1] = '.';
    cstr()[size - 2] = '.';
    cstr()[size - 3] = '.';
  }
}

class Widget::_Widget : public Fl_Menu_Bar
{
public:

  _Widget() :
    Fl_Menu_Bar(0, 0, 0, 0)
  {}
};

Group::Group()
{}

Group::Group(const Group & in)
{
  set(in);
}

Group::~Group()
{
  del();
}

Group & Group::operator = (const Group & in)
{
  if (&in != this) set(in);
  return *this;
}
  
void Group::set(const Group & in)
{
  del();
  
  _list.resize(in._list.size());
  for (uint i = 0; i < in._list.size(); ++i)
    _list[i] = new Item(*in._list[i]);
}

int Group::add(
  const Str & label,
  int shortcut,
  Fl_Callback * callback,
  void * data,
  int flags,
  bool value
) {
  _list += new Item(
    label,
    shortcut,
    callback,
    data,
    flags | (value ? VALUE : 0));

  return int(_list.size()) - 1;
}

List<int> Group::add(
  const List<Str> & label,
  const List<int> & shortcut,
  Fl_Callback * callback,
  void * data,
  int flags,
  int value
) {
  List<int> out;
  
  for (uint i = 0; i < label.size(); ++i)
    out += add(
      label_list(label[i], i), i < shortcut.size() ? shortcut[i] : 0,
      callback, data,
      flags, value != -1 ? (int(i) == value) : false);

  return out;
}

const List<Item *> & Group::get() const { return _list; }

void Group::end()
{
  _list += new Item;
}

void Group::del()
{
  for (uint i = 0; i < _list.size(); ++i)
    delete _list[i];
  _list.clear();
}

void Group::fl(mem::Buffer<Fl_Menu_Item> * out) const
{
  const uint size = _list.size();
  out->size(size + 1);
  out->zero();
    
  Fl_Menu_Item * p = (*out)();
  for (uint i = 0; i < size; ++i)
  {
    p[i].text = _list[i]->cstr();
    p[i].shortcut_ = _list[i]->shortcut;
    p[i].callback_ = _list[i]->callback;
    p[i].user_data_ = _list[i]->data;
    p[i].flags =
      (_list[i]->flags & TOGGLE ? FL_MENU_TOGGLE : 0) |
      (_list[i]->flags & RADIO ? FL_MENU_RADIO : 0) |
      (_list[i]->flags & VALUE ? FL_MENU_VALUE : 0) |
      (_list[i]->flags & SUB_MENU ? FL_SUBMENU : 0) |
      (_list[i]->flags & DIVIDER ? FL_MENU_DIVIDER : 0);
    //p[i].labelfont_ = (_list[i]->flags & SUB_MENU) ?
    //  style::global()->font_bold() : 0;
  }
}

Widget::Widget() :
  _test_shortcut(false)
{
  edges(false, false, false, true);
  
  _widget = new _Widget;
  _widget->box(FL_FLAT_BOX);
  Frame::add(_widget);
  
  layout::Item::size(layout::STRETCH, layout::FIXED);
}

void Widget::set(const menu::Group & in)
{
  _group = in;
  
  mem::Buffer<Fl_Menu_Item> tmp;
  _group.fl(&tmp);
  _widget->clear();
  _widget->copy(tmp());
}

const menu::Group & Widget::get() const { return _group; }

int Widget::item() const { return _widget->value(); }

bool Widget::value() const
{
  DJV_DBG_OBJ("Widget::value");
  
  //return _widget->mvalue()->value();

  const int item = _widget->value();
  bool value = _widget->menu()[item].value() & FL_MENU_VALUE;
  if (_test_shortcut) value = ! value;
  DJV_DBG_PRN("item = " << item);
  DJV_DBG_PRN("value = " << value);
  
  return value;
}

int Widget::test_shortcut()
{
  DJV_DBG_OBJ("Widget::test_shortcut");
  
  const List<menu::Item *> & list = _group.get();
  
  for (uint i = 0; i < list.size(); ++i)
    if (Fl::test_shortcut(list[i]->shortcut))
    {
      DJV_DBG_PRN("item = " << i);
      
      _widget->value(i);
      
      const Fl_Menu_Item & item = _widget->menu()[i];
      if (item.callback())
      {
        _test_shortcut = true;
        item.do_callback(_widget);
        _test_shortcut = false;
      }
      
      return 1;
    }
  
  return 0;
}

void Widget::dirty()
{
  Frame::dirty();

  _widget->textfont(text_font());
  _widget->textsize(style::global()->font_size());
  _widget->textcolor(text_color());
  
  fl_font(text_font(), style::global()->font_size());
  size_hint(Vec2i(0, fl_height()) + style::global()->margin_widget() * 2);
}

void Widget::geom(const Box2i & in)
{
  Frame::geom(in);
  _widget->resize(in.x, in.y, in.w, in.h);
}

const Box2i & Widget::geom() const { return Frame::geom(); }

}
}
