/*------------------------------------------------------------------------------
 lib/djv_ui/float.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include <djv_ui/button.h>
#include <djv_ui/label.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/num_float.h>
#include <djv_ui/slider.h>
#include <djv_ui/text.h>
#include <FL/Fl.H>

namespace djv_ui {
namespace num {

struct Float_Edit::_Float_Edit
{
  text::Edit * widget;
};

Float_Edit::Float_Edit() :
  sig(this),
  _p(new _Float_Edit)
{
  // Widgets.
  
  _p->widget = new text::Edit;
  _p->widget->size_str(format_float);
  
  Inc * inc_widget = new Inc;
  
  // Layout.
  
  size(layout::FIXED, layout::FIXED);
  align(layout::LEFT, layout::CENTER);
  
  Layout_H * layout = new Layout_H(this);
  layout->spacing(2);
  layout->margin(0);
  layout->add(_p->widget);
  layout->stretch(_p->widget);
  layout->add(inc_widget);

  // Initialize.
  
  value_update();

  // Callbacks.
  
  _p->widget->sig.set(this, widget_callback);
  inc_widget->inc_sig.set(this, inc_callback);
  inc_widget->dec_sig.set(this, dec_callback);
}

Float_Edit::~Float_Edit()
{}

int Float_Edit::handle(int in)
{
  if (Float_Base::handle(in)) return 1;
  return widget::Widget::handle(in);
}

layout::Item * Float_Edit::group_align() { return _p->widget->group_align(); }

void Float_Edit::value_callback()
{
  value_update();
  
  sig.emit(get());
}

void Float_Edit::widget_callback(const Str & in)
{
  set(str::str_to_float(in));
  
  value_callback();
}

void Float_Edit::inc_callback(bool) { inc_small_action(); }
void Float_Edit::dec_callback(bool) { dec_small_action(); }

void Float_Edit::value_update()
{
  callbacks(false);
  
  _p->widget->set(str::float_to_str(get(), precision()));
  
  callbacks(true);
}

struct Float_Display::_Float_Display
{
  text::Display * widget;
};

Float_Display::Float_Display() :
  _precision(0),
  _p(new _Float_Display)
{
  // Widgets.
  
  _p->widget = new text::Display;
  _p->widget->size_str(format_float);
  
  // Layout.
  
  size(layout::FIXED, layout::FIXED);
  align(layout::LEFT, layout::CENTER);
  
  Layout_H * layout = new Layout_H(this);
  layout->spacing(0);
  layout->margin(0);
  layout->add(_p->widget);
}

Float_Display::~Float_Display()
{}

void Float_Display::set(double in)
{
  _p->widget->set(str::float_to_str(in, _precision));
}

void Float_Display::precision(int in)
{
  _precision = in;
}

int Float_Display::precision() const { return _precision; }

void Float_Display::text_color(const Fl_Color & in)
{
  widget::Widget::text_color(in);
  _p->widget->text_color(in);
}

const Fl_Color & Float_Display::text_color() const
  { return widget::Widget::text_color(); }

void Float_Display::draw()
{
  widget::Widget::draw();
}

void Float_Display::dirty()
{
  widget::Widget::dirty();
  _p->widget->text_color(text_color());
}

struct Float_Slider::_Float_Slider
{
  _Float_Slider() :
    label(0)
  {}

  text::Edit * widget;
  slider::Float * slider;
  widget::Widget * label;
  button::Tool * _default;
  layout::Horizontal * layout;
};

Float_Slider::Float_Slider(double min, double max,
  const Str & label) :
  Float_Base(min, max),
  sig(this),
  _p(new _Float_Slider)
{
  clamp(false);
  
  // Widgets.
  
  _p->widget = new text::Edit;
  _p->widget->size_str(format_float);
  
  Inc * inc_widget = new Inc;

  _p->slider = new slider::Float;
  
  if (! label.empty())
    _p->label = new label::Widget(label);
  
  _p->_default = new button::Tool("tool_x");
  _p->_default->tooltip(base::tooltip_reset);
  _p->_default->deactivate();
  _p->_default->hide();
  
  // Layout.
  
  size(layout::STRETCH, layout::FIXED);
  
  _p->layout = new Layout_H(this);
  _p->layout->spacing(2);
  _p->layout->margin(0);
  if (_p->label)
    _p->layout->add(_p->label);
  _p->layout->add(_p->widget);
  _p->layout->add(inc_widget);
  _p->layout->add(_p->slider);
  _p->layout->stretch(_p->slider);
  _p->layout->add(_p->_default);

  // Initialize.
  
  range_update();
  value_update();

  // Callbacks.
  
  _p->widget->sig.set(this, widget_callback);
  inc_widget->inc_sig.set(this, inc_callback);
  inc_widget->dec_sig.set(this, dec_callback);
  _p->slider->sig.set(this, slider_callback);
  _p->_default->sig.set(this, default_callback);
}

Float_Slider::~Float_Slider()
{}

void Float_Slider::default_valid(bool in)
{
  Float_Base::default_valid(in);
  
  if (in) _p->_default->show();
  else _p->_default->hide();
  update();
}

bool Float_Slider::default_valid() const
  { return Float_Base::default_valid(); }

void Float_Slider::range(double min, double max)
{
  Float_Base::range(min, max);
  slider_range(min, max);
}

void Float_Slider::slider_range(double min, double max)
{
  _p->slider->range(min, max);
}

void Float_Slider::label(widget::Widget * in)
{
  if (_p->label) _p->layout->del(_p->label);
  _p->label = in;
  if (_p->label) _p->layout->insert(_p->label, 0);
  _p->layout->stretch(_p->slider);
  
  update();
}

widget::Widget * Float_Slider::label() const { return _p->label; }
layout::Item * Float_Slider::label_align() { return _p->label; }

void Float_Slider::text_font(const Fl_Font & in)
{
  widget::Widget::text_font(in);
  if (_p->label) _p->label->text_font(in);
}

const Fl_Font & Float_Slider::text_font() const
  { return widget::Widget::text_font(); }

void Float_Slider::text_color(const Fl_Color & in)
{
  widget::Widget::text_color(in);
  if (_p->label) _p->label->text_color(in);
}

const Fl_Color & Float_Slider::text_color() const
  { return widget::Widget::text_color(); }

int Float_Slider::handle(int in)
{
  if (Float_Base::handle(in)) return 1;
  return widget::Widget::handle(in);
}

layout::Item * Float_Slider::group_align()
  { return _p->widget->group_align(); }

void Float_Slider::value_callback()
{
  value_update();
  
  sig.emit(get());
}

void Float_Slider::range_callback()
{
  range_update();
}

void Float_Slider::widget_callback(const Str & in)
{
  Float_Base::set(str::str_to_float(in));
  
  value_callback();
}

void Float_Slider::inc_callback(bool)
{
  inc_small_action();
}

void Float_Slider::dec_callback(bool)
{
  dec_small_action();
}

void Float_Slider::slider_callback(double in)
{
  Float_Base::set(in);
}

void Float_Slider::default_callback(bool)
{
  Float_Base::set(default_value());
}

void Float_Slider::value_update()
{
  callbacks(false);
  
  const double value = get();
  
  _p->widget->set(str::float_to_str(value));
  _p->slider->set(value);
  
  if (default_valid() && value != default_value())
    _p->_default->activate();
  else
    _p->_default->deactivate();

  callbacks(true);
}

void Float_Slider::range_update()
{
  callbacks(false);
  
  _p->slider->range(min(), max());
  _p->slider->inc(inc_small(), inc_large());
  
  callbacks(true);
}

const Str format_float = "000.00";

}
}

