/*------------------------------------------------------------------------------
 plugin/djv_png/save.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "io.h"

namespace djv_png {
namespace io {

extern "C" {
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() :
  _f(0),
  _png(0),
  _png_info(0)
{}

Save::~Save()
{
  _close();
}

plugin::Plugin * Save::copy() const { return new Save; }

const Str & Save::name() const { return djv_png::name; }

const List<Str> & Save::extension_list() const
  { return djv_png::extension_list; }

void Save::open(const File & in, const Info & info, const Tag &,
  const djv_image::Time & time) throw (Error)
{
  DJV_DBG_OBJ("Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  if (time.list.size() > 1) _file.type = file::SEQ;
  
  _info = Info();
  _info.size = info.size;
  pixel::TYPE type = pixel::type(info.pixel);
  switch (type)
  {
    case pixel::U10:
    case pixel::F16: 
    case pixel::F32: type = pixel::U16; break;
    default: break;
  }
  _info.pixel = pixel::pixel(pixel::format(info.pixel), type);
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

namespace {

bool open(FILE * f, png_structp png, png_infop * png_info, const Info & info,
  const Tag & tag)
{
  if (setjmp(png_jmpbuf(png))) return false;

  *png_info = png_create_info_struct(png);
  if (! *png_info) return false;
  
  png_init_io(png, f);
  
  int color_type = 0;
  switch (info.pixel)
  {
    case pixel::L_U8:
    case pixel::L_U16: color_type = PNG_COLOR_TYPE_GRAY; break;
    case pixel::LA_U8:
    case pixel::LA_U16: color_type = PNG_COLOR_TYPE_GRAY_ALPHA; break;
    case pixel::RGB_U8:
    case pixel::RGB_U16: color_type = PNG_COLOR_TYPE_RGB; break;
    case pixel::RGBA_U8:
    case pixel::RGBA_U16: color_type = PNG_COLOR_TYPE_RGB_ALPHA; break;
    default: break;
  }
  
  png_set_IHDR(
    png, *png_info,
    info.size.x, info.size.y,
    pixel::bit_depth(info.pixel), color_type,
    PNG_INTERLACE_NONE,
    PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
  
  png_write_info(png, *png_info);
  
  return true;
}

}

void Save::_open(const File & in, const Tag & tag) throw (Error)
{
  DJV_DBG_OBJ("Save::_open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Initialize libpng.
  
  _png = png_create_write_struct(
    PNG_LIBPNG_VER_STRING, 0, djv_png_error, djv_png_warning);
  if (! _png)
    throw Error(name(), _png_error.msg);

  _png->error_ptr = &_png_error;

  SNPRINTF(_png_error.msg, str::cstr_size, djv_image::io::error_open.c_str());
  
  // Open.
  
  _f = ::fopen(in.get().c_str(), "wb");
  if (! _f)
    throw Error(name(), djv_image::io::error_open);

  if (! io::open(_f, _png, &_png_info, _image.info(), tag))
    throw Error(name(), _png_error.msg);
}

void Save::_close()
{
  if (_png || _png_info)
  {
    png_destroy_write_struct(
      _png ? &_png : 0,
      _png_info ? &_png_info : 0);
    _png = 0;
    _png_info = 0;
  }
  
  if (_f) { ::fclose(_f); _f = 0; }
}

namespace {

bool scanline(png_structp png, const uint8_t * in)
{
  if (setjmp(png_jmpbuf(png))) return false;
  png_write_row(png, (png_byte *)in);
  return true;
}

bool end(png_structp png, png_infop png_info)
{
  if (setjmp(png_jmpbuf(png))) return false;
  png_write_end(png, png_info);
  return true;
}

}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ("Save::set");
  DJV_DBG_PRN("in = " << in);

  // Open file.
  
  _open(_file.get(frame), in.tag);
  
  // Temporary buffers.
  
  const Data * p = &in;
  if (in.info() != _image.info())
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl_data::copy(in, &_image);
    p = &_image;
  }
  
  // Write.
  
  const int h = p->h();
  for (int y = 0; y < h; ++y)
    if (! scanline(_png, p->data(0, h - 1 - y)))
      throw Error(name(), _png_error.msg);

  if (! end(_png, _png_info))
    throw Error(name(), _png_error.msg);

  _close();
}

}
}

