/*------------------------------------------------------------------------------
 plugin/djv_quicktime_unix/save.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "io.h"

namespace djv_quicktime {
namespace io {

extern "C" {

DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save()
{
  quicktime_init();
  
  return new Save;
}

}

Save::Save() :
  _f(0),
  _frame(0),
  _frame_duration(0)
{}

plugin::Plugin * Save::copy() const
{
  Save * out = new Save;
  out->_option = _option;
  return out;
}

const Str & Save::name() const { return djv_quicktime::name; }

const List<Str> & Save::extension_list() const
  { return djv_quicktime::extension_list; }

bool Save::option(const Str & in, Str * data)
{
try
{
  if (str::compare(in, option_list()[SAVE_CODEC])) *data >> _option.codec;
}
catch (Str) { return false; }
  return true;
}

Str Save::option(const Str & in) const
{
  Str out;
  if (str::compare(in, option_list()[SAVE_CODEC])) out << _option.codec;
  return out;
}

const List<Str> & Save::option_list() const { return label_save_option(); }

namespace {

struct Codec_Info
{
  Codec_Info(const Str & name) :
    p(lqt_find_video_codec_by_name(name.c_str()))
  {}

  ~Codec_Info()
  {
    if (p) lqt_destroy_codec_info(p);
  }
    
  lqt_codec_info_t ** p;
};

}

void Save::open(const File & in, const Info & info, const Tag & tag,
  const djv_image::Time & time) throw (Error)
{
  DJV_DBG_OBJ("Save::open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("info = " << info);
  
  // Open.
  
  _file = in;
  
  _f = lqt_open_write(in.get().c_str(), LQT_FILE_QT);
  
  if (! _f)
    throw Error(name(), djv_image::io::error_open);
  
  // Information.
  
  Codec_Info codec_info(_option.codec);
  
  if (! codec_info.p)
    throw Error(name(), error_codec);
  
  _frame = 0;
  _frame_duration = time.speed.duration;
  
  // XXX Return value?

  lqt_set_video(_f, 1, info.size.x, info.size.y, _frame_duration,
    time.speed.scale, codec_info.p[0]);
  
  _info = Info();
  _info.size = info.size;
  _info.pixel = info.pixel;
  
  static int cmodel_list [] =
  {
    BC_RGB888,
    BC_BGR888,
    BC_RGB161616,
    BC_RGBA8888,
    BC_RGBA16161616,
    LQT_COLORMODEL_NONE
  };
  const int cmodel = lqt_get_best_colormodel(_f, 0, cmodel_list);
  switch (cmodel)
  {
    case BC_BGR888: _info.bgr = true;
    case BC_RGB888: _info.pixel = pixel::RGB_U8; break;
    case BC_RGB161616: _info.pixel = pixel::RGB_U16; break;
    case BC_RGBA8888: _info.pixel = pixel::RGBA_U8; break;
    case BC_RGBA16161616: _info.pixel = pixel::RGBA_U16; break;
    default: throw Error(name(), error_colormodel);
  }
  lqt_set_cmodel(_f, 0, cmodel);
  DJV_DBG_PRN("info = " << _info);
  
  const List<Str> & label_tag = tag::label_tag();
  Str _tag = tag[label_tag[tag::CREATOR]];
  if (_tag.size()) lqt_set_artist(_f, (char *)_tag.c_str());
  _tag = tag[label_tag[tag::DESCRIPTION]];
  if (_tag.size()) lqt_set_comment(_f, (char *)_tag.c_str());
  _tag = tag[label_tag[tag::COPYRIGHT]];
  if (_tag.size()) quicktime_set_copyright(_f, (char *)_tag.c_str());

  _image.set(_info);
}

void Save::close() throw (Error)
{
  DJV_DBG_OBJ("Save::close");

  if (_f) { quicktime_close(_f); _f = 0; }
}

namespace {

struct Rows
{
  Rows(const Data * in) :
    p(0)
  {
    p = new unsigned char * [in->h() + 1];
    for (int y = 0; y < in->h(); ++y)
      p[y] = const_cast<pixel::U8_T *>(in->data(0, in->h() - 1 - y));
  }
    
  ~Rows()
  {
    if (p) delete [] p;
  }
  
  unsigned char ** p;
};

}

void Save::set(const Image & in, int) throw (Error)
{
  DJV_DBG_OBJ("Save::set");
  DJV_DBG_PRN("in = " << in);

  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl_data::copy(in, &_image);
    p = &_image;
  }

  lqt_encode_video(_f, Rows(p).p, 0, _frame * _frame_duration);
  
  ++_frame;
}

const List<Str> & label_save_option()
{
  static const List<Str> data = List<Str>() <<
    "Codec";
  return data;
}

}
}

